# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from dataclasses import dataclass
from lxml import etree  # nosec # We only write XML files

from ...data_model.container import CoverageContainer
from ...data_model.stats import CoverageStat, SummarizedStats
from ...options import Options
from ...utils import write_xml_output


def write_report(
    covdata: CoverageContainer, output_file: str, options: Options
) -> None:
    """produce an XML report in the JaCoCo format"""

    root_elem = etree.Element("report")
    root_elem.set("name", options.jacoco_report_name)

    # Generate the coverage output (on a per-package basis)
    packages = dict[str, PackageData]()

    for _, filecov in sorted(covdata.items()):
        filename = filecov.presentable_filename(options.root_filter)
        if "/" in filename:
            directory, fname = filename.rsplit("/", 1)
        else:
            directory, fname = "", filename

        package_data = packages.setdefault(
            directory,
            PackageData(
                {},
                SummarizedStats.new_empty(),
            ),
        )
        source_elem = etree.Element("sourcefile")
        source_elem.set("name", fname)

        for linecov in filecov.linecov(sort=True):
            line_elem = etree.SubElement(source_elem, "line")
            line_elem.set("nr", str(linecov.lineno))
            if linecov.is_reportable:
                stat = linecov.branch_coverage()
                if stat.total:
                    line_elem.set("mb", str(stat.total - stat.covered))
                    line_elem.set("cb", str(stat.covered))

        filecov_stats = filecov.stats
        add_counters(source_elem, filecov_stats)

        package_data.sources[fname] = source_elem
        package_data.stats += filecov_stats

    for package_name, package_data in sorted(packages.items()):
        package_elem = etree.SubElement(root_elem, "package")
        package_elem.set("name", package_name.replace("/", "."))
        for _, source_data in sorted(package_data.sources.items()):
            package_elem.append(source_data)
        add_counters(package_elem, package_data.stats)

    add_counters(root_elem, covdata.stats)

    write_xml_output(
        root_elem,
        pretty=options.jacoco_pretty,
        filename=output_file,
        default_filename="jacoco.xml",
        doctype="<!DOCTYPE report SYSTEM 'https://www.jacoco.org/jacoco/trunk/coverage/report.dtd'>",
    )


@dataclass
class PackageData:
    """Class holding package information."""

    sources: dict[str, etree._Element]
    stats: SummarizedStats


def add_counters(elem: etree._Element, stats: SummarizedStats) -> None:
    """Add the counter elements for the given stats."""

    def add_counter_element(element_type: str, stat: CoverageStat) -> None:
        """Add one stat element."""
        counter_elem = etree.SubElement(elem, "counter")
        counter_elem.set("type", element_type)
        counter_elem.set("missed", str(stat.total - stat.covered))
        counter_elem.set("covered", str(stat.covered))

    add_counter_element("LINE", stats.line)
    add_counter_element("BRANCH", stats.branch)
    add_counter_element("METHOD", stats.function)



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

# cspell:ignore xmlcharrefreplace

import functools
import os
import re
from typing import Any, Callable, Iterable, Iterator

from jinja2 import (
    BaseLoader,
    Environment,
    ChoiceLoader,
    FileSystemLoader,
    FunctionLoader,
    PackageLoader,
    Template,
)
from markupsafe import Markup
import pygments
from pygments.filter import Filter
from pygments.formatters.html import HtmlFormatter
from pygments.lexer import Lexer
from pygments.lexers import get_lexer_for_filename
from pygments.token import _TokenType, Token
from pygments.style import Style
from pygments.styles.default import DefaultStyle

from ...data_model.container import CoverageContainer, CoverageContainerDirectory
from ...data_model.coverage import (
    DecisionCoverageConditional,
    DecisionCoverageSwitch,
    DecisionCoverageUncheckable,
    FileCoverage,
    LineCoverage,
    CoverageStat,
    DecisionCoverageStat,
)
from ...data_model.coverage_dict import FunctioncovKeyType
from ...exclusions.markers import _EXCLUDE_FLAG, get_markers_regex
from ...logging import LOGGER
from ...options import Options
from ...utils import (
    GZIP_SUFFIX,
    chdir,
    commonpath,
    force_unix_separator,
    get_md5_hexdigest,
    get_version_for_report,
    open_text_for_writing,
)


PYGMENTS_CSS_MARKER = "/* Comment.Preproc */"


# html_theme string is <theme_directory>.<color> or only <color> (if only color use default)
# examples: github.green github.blue or blue or green
def get_theme_name(html_theme: str) -> str:
    """Get the theme name without the color."""
    return html_theme.split(".")[0] if "." in html_theme else "default"


def get_theme_color(html_theme: str) -> str:
    """Get the theme color from the theme name."""
    return html_theme.split(".")[1] if "." in html_theme else html_theme


@functools.lru_cache(maxsize=1)
def templates(options: Options) -> Environment:
    """Get the Jinja2 environment for the templates."""
    # As default use the package loader
    loaders: list[BaseLoader] = []
    # If a directory is given files in the directory have higher precedence.
    if options.html_template_dir is not None:
        loaders.append(FileSystemLoader(options.html_template_dir))

    loaders += [
        PackageLoader(
            "gcovr.formats.html",
            package_path=get_theme_name(options.html_theme),
        ),
        PackageLoader(
            "gcovr.formats.html",
            package_path="common",
        ),
    ]

    return Environment(
        loader=ChoiceLoader(loaders),
        autoescape=True,
        trim_blocks=True,
        lstrip_blocks=True,
    )


@functools.lru_cache(maxsize=1)
def user_templates() -> Environment:
    """Get the Jinja2 environment for the user templates."""

    def load_user_template(template: str) -> str | None:
        contents = None
        try:
            with open(template, "rb") as f:
                contents = f.read().decode("UTF-8")
        # This exception can only occur if the file gets inaccessible while gcovr is running.
        except FileNotFoundError:  # pragma: no cover
            pass

        return contents

    return Environment(
        loader=FunctionLoader(load_user_template),
        autoescape=True,
        trim_blocks=True,
        lstrip_blocks=True,
    )


class CssRenderer:
    """Class for rendering the CSS template with Jinja2."""

    @staticmethod
    def __load_css_template(options: Options) -> Template:
        """Load the CSS template."""
        if options.html_css is not None:
            template_path = os.path.relpath(options.html_css)
            return user_templates().get_template(template_path)

        return templates(options).get_template("style.css")

    @staticmethod
    def render(options: Options) -> str:
        """Get the rendered CSS content."""
        template = CssRenderer.__load_css_template(options)
        return template.render(tab_size=options.html_tab_size)


class NullHighlighting:
    """Class if no syntax highlighting is available for the given file."""

    def get_css(self) -> str:
        """Get the empty CSS."""
        return ""

    @staticmethod
    def highlighter_for_file(_: str) -> Callable[[str], list[str]]:
        """Get the default highlighter which only returns the content as raw text lines."""
        return lambda code: [line.rstrip() for line in code.split("\n")]


class PygmentsHighlighting:
    """Class for syntax highlighting in report."""

    class DefaultStyle(Style):
        """GCOVR default style."""

        styles = dict(
            list(DefaultStyle.styles.items()) + [(Token.Comment.Special, "bold")]
        )

    class MarkerFilter(Filter):
        """A filter to highlight the marker keywords"""

        def __init__(self, markers_regex: re.Pattern[str], **options: Any):
            super().__init__(**options)
            self.markers_regex = markers_regex

        def filter(
            self, lexer: Lexer, stream: Iterable[tuple[_TokenType, str]]
        ) -> Iterator[tuple[_TokenType, str]]:
            for ttype, value in stream:
                if _EXCLUDE_FLAG in value:
                    last = 0
                    for match in self.markers_regex.finditer(value):
                        start, end = match.start(), match.end()
                        if start != last:
                            yield ttype, value[last:start]
                        yield Token.Comment.Special, value[start:end]
                        last = end
                    if last != len(value):
                        yield ttype, value[last:]
                else:
                    yield ttype, value

    def __init__(self, style: str, markers_regex: re.Pattern[str]) -> None:
        self.filter = PygmentsHighlighting.MarkerFilter(markers_regex)
        self.formatter = None
        try:
            self.formatter = HtmlFormatter(
                nowrap=True,
                style=PygmentsHighlighting.DefaultStyle
                if style == "default"
                else style,
            )
        except ImportError as e:  # pragma: no cover
            LOGGER.warning("No syntax highlighting available: %s", str(e))

    def get_css(self) -> str:
        """Get the CSS for the syntax highlighting."""
        if self.formatter is None:  # pragma: no cover
            return ""
        return (
            f"\n\n/* pygments syntax highlighting */\n{self.formatter.get_style_defs()}"  # type: ignore [no-untyped-call]
        )

    def highlighter_for_file(self, filename: str) -> Callable[[str], list[str]]:
        """Get the highlighter for the given filename."""
        if self.formatter is None:  # pragma: no cover
            return NullHighlighting.highlighter_for_file(filename)

        try:
            lexer = get_lexer_for_filename(filename, None, stripnl=False)
            lexer.add_filter(self.filter)
            formatter = self.formatter
            return lambda code: [
                Markup(line.rstrip())  # nosec
                for line in pygments.highlight(code, lexer, formatter).split("\n")
            ]
        except pygments.util.ClassNotFound:  # pragma: no cover
            return NullHighlighting.highlighter_for_file(filename)


@functools.lru_cache(maxsize=1)
def get_formatter(options: Options) -> PygmentsHighlighting | NullHighlighting:
    """Get the formatter for the selected theme."""
    if options.html_syntax_highlighting:
        highlight_style = (
            templates(options)
            .get_template(f"pygments.{get_theme_color(options.html_theme)}")
            .render()
        )
        return PygmentsHighlighting(
            highlight_style, get_markers_regex(options.exclude_pattern_prefix)
        )

    return NullHighlighting()


def coverage_to_class(
    coverage: float | None, medium_threshold: float, high_threshold: float
) -> str:
    """Get the coverage class depending on the threshold."""
    if coverage is None:
        return "coverage-unknown"
    if coverage == 0:
        return "coverage-none"
    if coverage < medium_threshold:
        return "coverage-low"
    if coverage < high_threshold:
        return "coverage-medium"
    return "coverage-high"


class RootInfo:
    """Class holding the information used in Jinja2 template."""

    def __init__(self, options: Options) -> None:
        self.sort_by = (
            "filename"
            if options.sort_key == "filename"
            else ("branches" if options.sort_branches else "lines")
        )
        self.sort_percent = options.sort_key != "uncovered-number"
        self.sorted = (
            "sorted-descending" if options.sort_reverse else "sorted-ascending"
        )
        self.medium_threshold = options.medium_threshold
        self.high_threshold = options.high_threshold
        self.medium_threshold_line = options.medium_threshold_line
        self.high_threshold_line = options.high_threshold_line
        self.medium_threshold_branch = options.medium_threshold_branch
        self.high_threshold_branch = options.high_threshold_branch
        self.link_function_list = (
            options.html_details or options.html_nested
        ) and not (options.html_single_page and options.html_static_report)
        self.relative_anchors = options.html_relative_anchors
        self.single_page = options.html_single_page
        self.static_report = options.html_static_report

        self.version = get_version_for_report()
        self.head = options.html_title
        self.date = options.timestamp.isoformat(sep=" ", timespec="seconds")
        self.encoding = options.html_encoding
        self.directory = ""
        self.branches = dict[str, Any]()
        self.conditions = dict[str, Any]()
        self.decisions = dict[str, Any]()
        self.calls = dict[str, Any]()
        self.functions = dict[str, Any]()
        self.lines = dict[str, Any]()
        self.navigation = dict[str, tuple[str | None, str | None]]()

    def set_directory(self, directory: str) -> None:
        """Set the directory for the report."""
        self.directory = directory

    def get_directory(self) -> str:
        """Get the directory for the report."""
        return (
            "." if self.directory == "" else force_unix_separator(str(self.directory))
        )

    def set_coverage(self, covdata: CoverageContainer) -> None:
        """Update this RootInfo with a summary of the CoverageContainer."""
        stats = covdata.stats
        self.lines = dict_from_stat(stats.line, self.line_coverage_class, 0.0)
        self.functions = dict_from_stat(stats.function, self.coverage_class)
        self.branches = dict_from_stat(stats.branch, self.branch_coverage_class)
        self.conditions = dict_from_stat(stats.condition, self.coverage_class)
        self.decisions = dict_from_stat(stats.decision, self.coverage_class)
        self.calls = dict_from_stat(stats.call, self.coverage_class)

    def line_coverage_class(self, coverage: float | None) -> str:
        """Get the coverage class for the line."""
        return coverage_to_class(
            coverage, self.medium_threshold_line, self.high_threshold_line
        )

    def branch_coverage_class(self, coverage: float | None) -> str:
        """Get the coverage class for the branch."""
        return coverage_to_class(
            coverage, self.medium_threshold_branch, self.high_threshold_branch
        )

    def coverage_class(self, coverage: float | None) -> str:
        """Get the coverage class for all other types."""
        return coverage_to_class(coverage, self.medium_threshold, self.high_threshold)


#
# Produce an HTML report
#
def write_report(
    covdata: CoverageContainer, output_file: str, options: Options
) -> None:
    """Write the HTML report"""
    css_data = CssRenderer.render(options).strip()
    medium_threshold = options.medium_threshold
    high_threshold = options.high_threshold
    medium_threshold_line = options.medium_threshold_line
    high_threshold_line = options.high_threshold_line
    medium_threshold_branch = options.medium_threshold_branch
    high_threshold_branch = options.high_threshold_branch
    show_calls = options.show_calls
    show_decision = options.show_decision

    data = dict[str, Any]()
    root_info = RootInfo(options)
    data["info"] = root_info

    data["SHOW_DECISION"] = show_decision
    data["SHOW_CALLS"] = show_calls
    data["SHOW_CONDITION_COVERAGE"] = any(
        filter(lambda filecov: filecov.condition_coverage().total > 0, covdata.values())  # type: ignore [arg-type]
    )
    data["USE_BLOCK_IDS"] = options.html_block_ids
    data["COVERAGE_MED"] = medium_threshold
    data["COVERAGE_HIGH"] = high_threshold
    data["LINE_COVERAGE_MED"] = medium_threshold_line
    data["LINE_COVERAGE_HIGH"] = high_threshold_line
    data["BRANCH_COVERAGE_MED"] = medium_threshold_branch
    data["BRANCH_COVERAGE_HIGH"] = high_threshold_branch

    self_contained = options.html_self_contained
    if self_contained is None:
        self_contained = (
            not (options.html_details or options.html_nested)
            or options.html_single_page
        )

    if output_file.endswith(os.sep):
        if options.html_single_page:
            output_file += "coverage_single_page.html"
        elif options.html_nested:
            output_file += "coverage_nested.html"
        elif options.html_details:
            output_file += "coverage_details.html"
        else:
            output_file += "coverage.html"

    if PYGMENTS_CSS_MARKER in css_data:
        LOGGER.info(
            "Skip adding of pygments styles since %r found in user stylesheet",
            PYGMENTS_CSS_MARKER,
        )
    else:
        css_data += get_formatter(options).get_css()

    if options.html_details or options.html_nested:
        data["ROOT_FNAME"] = os.path.basename(output_file)
        (output_prefix, output_suffix) = _get_prefix_and_suffix(output_file)
        functions_output_file = f"{output_prefix}.functions{output_suffix}"
        data["FUNCTIONS_FNAME"] = os.path.basename(functions_output_file)
        if options.html_single_page:
            # Remove the prefix to get shorter links
            data["FUNCTIONS_FNAME"] = data["FUNCTIONS_FNAME"].split(".", maxsplit=1)[1]

    javascript_data = (
        None
        if options.html_static_report
        else templates(options).get_template("gcovr.js").render(**data).strip()
    )

    if self_contained:
        data["css"] = css_data
        if javascript_data is not None:
            data["javascript"] = javascript_data
    else:
        css_output = os.path.splitext(output_file)[0] + ".css"
        with open_text_for_writing(css_output, encoding="utf-8") as fh_out:
            fh_out.write(css_data)
            fh_out.write("\n")

        if options.html_relative_anchors:
            css_link = os.path.basename(css_output)
        else:  # pragma: no cover  Can't be checked because of the reference compare
            css_link = css_output
        data["css_link"] = css_link

        if javascript_data is not None:
            javascript_output = os.path.splitext(output_file)[0] + ".js"
            with open_text_for_writing(javascript_output) as fh_out:
                fh_out.write(javascript_data)
                fh_out.write("\n")

            if options.html_relative_anchors:
                javascript_link = os.path.basename(javascript_output)
            else:  # pragma: no cover  Can't be checked because of the reference compare
                javascript_link = javascript_output
            data["javascript_link"] = javascript_link

    data["theme"] = get_theme_color(options.html_theme)

    root_info.set_coverage(covdata)
    if root_info.link_function_list and root_info.functions["total"] == 0:
        root_info.link_function_list = False

    # Generate the coverage output (on a per-package basis)
    # source_dirs = set()
    files = []
    filtered_fname = ""
    sorted_keys = covdata.sort_coverage(
        sort_key=options.sort_key,
        sort_reverse=options.sort_reverse,
        by_metric="branch" if options.sort_branches else "line",
        filename_uses_relative_pathname=True,
    )

    if options.html_nested:
        covdata.populate_directories(sorted_keys, options.root_filter)

    cdata_fname = dict[str, str]()
    cdata_sourcefile = dict[str, str | None]()
    for fname in sorted_keys + [v.dirname for v in covdata.directories]:
        filtered_fname = options.root_filter.sub("", fname)
        if filtered_fname != "":
            files.append(filtered_fname)
        cdata_fname[fname] = force_unix_separator(filtered_fname)
        if options.html_details or options.html_nested or options.html_single_page:
            if os.path.normpath(fname) == os.path.normpath(options.root_dir):
                cdata_sourcefile[fname] = (
                    output_file[: -len(GZIP_SUFFIX)]
                    if output_file.endswith(GZIP_SUFFIX)
                    else output_file
                )
            else:
                cdata_sourcefile[fname] = _make_short_source_filename(
                    output_file, filtered_fname.rstrip(os.sep)
                )
                if options.html_single_page and cdata_sourcefile[fname] is not None:
                    # Remove the prefix to get shorter links
                    cdata_sourcefile[fname] = str(cdata_sourcefile[fname]).split(
                        ".", maxsplit=1
                    )[1]
        else:
            cdata_sourcefile[fname] = None

    # Define the common root directory, which may differ from options.root_dir
    # when source files share a common prefix.
    root_directory = ""
    if len(files) > 1:
        common_dir = commonpath(files)
        if common_dir != "":
            root_directory = common_dir
    else:
        directory, _ = os.path.split(filtered_fname)
        if directory != "":
            root_directory = str(directory) + os.sep

    previous_fname: str | None = None
    previous_link_report: str | None = None
    for fname in sorted(cdata_sourcefile.keys()):
        root_info.navigation[fname] = (previous_link_report, None)
        link_report = cdata_sourcefile[fname]
        if link_report is not None:
            if root_info.relative_anchors or root_info.single_page:
                link_report = os.path.basename(link_report)
        if previous_fname is not None:
            root_info.navigation[previous_fname] = (
                root_info.navigation[previous_fname][0],
                link_report,
            )
        previous_fname = fname
        previous_link_report = link_report

    root_info.set_directory(force_unix_separator(root_directory))

    if options.html_single_page:
        write_single_page(
            options,
            root_info,
            output_file,
            covdata,
            sorted_keys,
            cdata_fname,
            cdata_sourcefile,
            data,
        )
    else:
        if options.html_nested and covdata.directories:
            write_directory_pages(
                options,
                root_info,
                output_file,
                covdata,
                cdata_fname,
                cdata_sourcefile,
                data,
            )
        else:
            write_root_page(
                options,
                root_info,
                output_file,
                covdata,
                cdata_fname,
                cdata_sourcefile,
                data,
                sorted_keys,
            )
            if not options.html_details:
                return

        write_source_pages(
            options,
            root_info,
            functions_output_file,
            covdata,
            cdata_fname,
            cdata_sourcefile,
            data,
        )


def write_root_page(
    options: Options,
    root_info: RootInfo,
    output_file: str,
    covdata: CoverageContainer,
    cdata_fname: dict[str, str],
    cdata_sourcefile: dict[str, Any],
    data: dict[str, Any],
    sorted_keys: list[str],
) -> None:
    """Generate the root HTML file that contains the high level report."""
    files = []
    for fname in sorted_keys:
        files.append(
            get_coverage_data(
                root_info, covdata[fname], cdata_sourcefile[fname], cdata_fname[fname]
            )
        )

    html_string = (
        templates(options)
        .get_template("directory_page.html")
        .render(**data, entries=files)
    )
    with open_text_for_writing(
        output_file, encoding=options.html_encoding, errors="xmlcharrefreplace"
    ) as fh:
        fh.write(html_string + "\n")


def write_source_pages(
    options: Options,
    root_info: RootInfo,
    functions_output_file: str,
    covdata: CoverageContainer,
    cdata_fname: dict[str, str],
    cdata_sourcefile: dict[str, Any],
    data: dict[str, Any],
) -> None:
    """Write a page for each source file."""
    error_no_files_not_found = 0
    all_functions = {}
    for fname, filecov in sorted(covdata.items()):
        file_data, functions, file_not_found = get_file_data(
            options, root_info, fname, cdata_fname, cdata_sourcefile, filecov
        )
        all_functions.update(functions)
        if file_not_found:
            error_no_files_not_found += 1

        html_string = (
            templates(options)
            .get_template("source_page.html")
            .render(**data, **file_data)
        )
        with open_text_for_writing(
            cdata_sourcefile[fname],
            encoding=options.html_encoding,
            errors="xmlcharrefreplace",
        ) as fh:
            fh.write(html_string + "\n")

    html_string = (
        templates(options)
        .get_template("functions_page.html")
        .render(
            **data,
            all_functions=[all_functions[k] for k in sorted(all_functions)],
        )
    )
    with open_text_for_writing(
        functions_output_file,
        encoding=options.html_encoding,
        errors="xmlcharrefreplace",
    ) as fh:
        fh.write(html_string + "\n")

    if error_no_files_not_found != 0:
        raise RuntimeError(f"{error_no_files_not_found} source file(s) not found.")


def write_directory_pages(
    options: Options,
    root_info: RootInfo,
    output_file: str,
    covdata: CoverageContainer,
    cdata_fname: dict[str, str],
    cdata_sourcefile: dict[str, Any],
    data: dict[str, Any],
) -> None:
    """Write a page for each directory."""
    # The first directory is the shortest one --> This is the root dir
    root_key = next(iter(sorted([d.dirname for d in covdata.directories])))

    directory_data = {}
    for dircov in covdata.directories:
        directory_data = get_directory_data(
            options, root_info, cdata_fname, cdata_sourcefile, dircov
        )
        html_string = (
            templates(options)
            .get_template("directory_page.html")
            .render(**data, **directory_data)
        )
        filename = None
        if dircov.dirname in [root_key, ""]:
            filename = output_file
        elif dircov.dirname in cdata_sourcefile:
            filename = cdata_sourcefile[dircov.dirname]
        else:
            LOGGER.warning(
                "There's a subdirectory %r that there's no source files within it",
                dircov.dirname,
            )

        if filename:
            with open_text_for_writing(
                filename, encoding=options.html_encoding, errors="xmlcharrefreplace"
            ) as fh:
                fh.write(html_string + "\n")


def write_single_page(
    options: Options,
    root_info: RootInfo,
    output_file: str,
    covdata: CoverageContainer,
    sorted_keys: list[str],
    cdata_fname: dict[str, str],
    cdata_sourcefile: dict[str, Any],
    data: dict[str, Any],
) -> None:
    """Write a single page HTML report."""
    error_no_files_not_found = 0
    files = []
    all_functions = {}
    for fname, filecov in sorted(covdata.items()):
        file_data, functions, file_not_found = get_file_data(
            options, root_info, fname, cdata_fname, cdata_sourcefile, filecov
        )
        all_functions.update(functions)
        if file_not_found:
            error_no_files_not_found += 1

        files.append(file_data)

    all_files = []
    for fname in sorted_keys:
        all_files.append(
            get_coverage_data(
                root_info, covdata[fname], cdata_sourcefile[fname], cdata_fname[fname]
            )
        )
    directories = list[dict[str, Any]]([{"entries": all_files}])
    if not root_info.static_report:
        for dircov in covdata.directories:
            directories.append(
                get_directory_data(
                    options, root_info, cdata_fname, cdata_sourcefile, dircov
                )
            )
    if len(directories) == 1:
        directories[0]["dirname"] = "/"  # We need this to have a correct id in HTML.

    html_string = (
        templates(options)
        .get_template("single_page.html")
        .render(
            **data,
            files=files,
            directories=directories,
            all_functions=[all_functions[k] for k in sorted(all_functions)],
        )
    )
    with open_text_for_writing(
        output_file,
        encoding=options.html_encoding,
        errors="xmlcharrefreplace",
    ) as fh:
        fh.write(html_string + "\n")

    if error_no_files_not_found != 0:
        raise RuntimeError(f"{error_no_files_not_found} source file(s) not found.")


def get_coverage_data(
    root_info: RootInfo,
    cdata: CoverageContainerDirectory | FileCoverage,
    link_report: str,
    cdata_fname: str,
    relative_path: str = "",
) -> dict[str, Any]:
    """Get the coverage data"""

    medium_threshold = root_info.medium_threshold
    high_threshold = root_info.high_threshold
    medium_threshold_line = root_info.medium_threshold_line
    high_threshold_line = root_info.high_threshold_line
    medium_threshold_branch = root_info.medium_threshold_branch
    high_threshold_branch = root_info.high_threshold_branch

    def coverage_class(coverage: float | None) -> str:
        return coverage_to_class(coverage, medium_threshold, high_threshold)

    def line_coverage_class(coverage: float | None) -> str:
        return coverage_to_class(coverage, medium_threshold_line, high_threshold_line)

    def branch_coverage_class(coverage: float | None) -> str:
        return coverage_to_class(
            coverage, medium_threshold_branch, high_threshold_branch
        )

    def sort_value(coverage_stats: CoverageStat | DecisionCoverageStat) -> str:
        return str(
            coverage_stats.percent_or("-")
            if root_info.sort_percent
            else coverage_stats.total - coverage_stats.covered
        )

    stats = cdata.stats

    is_file_with_lines = isinstance(cdata, FileCoverage) and cdata.has_lines()
    lines = {
        "total": stats.line.total_with_excluded,
        "exec": stats.line.covered,
        "excluded": stats.line.excluded,
        "coverage": stats.line.percent_or(100.0 if is_file_with_lines else "-"),
        "class": line_coverage_class(
            stats.line.percent_or(100.0 if is_file_with_lines else None)
        ),
        "sort": sort_value(stats.line),
    }

    branches = {
        "total": stats.branch.total_with_excluded,
        "exec": stats.branch.covered,
        "excluded": stats.branch.excluded,
        "coverage": stats.branch.percent_or("-"),
        "class": branch_coverage_class(stats.branch.percent),
        "sort": sort_value(stats.branch),
    }

    conditions = {
        "total": stats.condition.total_with_excluded,
        "exec": stats.condition.covered,
        "excluded": stats.condition.excluded,
        "coverage": stats.condition.percent_or("-"),
        "class": branch_coverage_class(stats.condition.percent),
        "sort": sort_value(stats.condition),
    }

    decisions = {
        "total": stats.decision.total,
        "exec": stats.decision.covered,
        "unchecked": stats.decision.uncheckable,
        "coverage": stats.decision.percent_or("-"),
        "class": coverage_class(stats.decision.percent),
        "sort": sort_value(stats.decision),
    }

    functions = {
        "total": stats.function.total_with_excluded,
        "exec": stats.function.covered,
        "excluded": stats.function.excluded,
        "coverage": stats.function.percent_or("-"),
        "class": coverage_class(stats.function.percent),
        "sort": sort_value(stats.function),
    }

    calls = {
        "total": stats.call.total_with_excluded,
        "exec": stats.call.covered,
        "excluded": stats.call.excluded,
        "coverage": stats.call.percent_or("-"),
        "class": coverage_class(stats.call.percent),
        "sort": sort_value(stats.call),
    }
    display_filename = force_unix_separator(
        os.path.relpath(
            os.path.realpath(cdata_fname),
            os.path.realpath(
                os.path.join(root_info.directory, relative_path)
                if relative_path
                else root_info.directory
            ),
        )
    )

    if link_report is not None:
        if root_info.relative_anchors or root_info.single_page:
            link_report = os.path.basename(link_report)

    return {
        "filename": display_filename,
        "link": link_report,
        "lines": lines,
        "branches": branches,
        "conditions": conditions,
        "decisions": decisions,
        "functions": functions,
        "calls": calls,
    }


def get_directory_data(
    options: Options,
    root_info: RootInfo,
    cdata_fname: dict[str, str],
    cdata_sourcefile: dict[str, str],
    covdata_dir: CoverageContainerDirectory,
) -> dict[str, Any]:
    """Get the data for a directory to generate the HTML"""
    relative_path = cdata_fname[covdata_dir.dirname]
    if relative_path == ".":
        relative_path = ""
    elif relative_path.startswith(root_info.directory):
        relative_path = relative_path[len(root_info.directory) :]
    directory_data = dict[str, Any](
        {
            "dirname": (
                cdata_sourcefile[covdata_dir.dirname]
                if cdata_fname[covdata_dir.dirname]
                else "/"
            ),
            "relative_path": relative_path,
        }
    )

    sorted_keys = covdata_dir.sort_coverage(
        sort_key=options.sort_key,
        sort_reverse=options.sort_reverse,
        by_metric="branch" if options.sort_branches else "line",
        filename_uses_relative_pathname=True,
    )

    files = []
    for key in sorted_keys:
        fname = covdata_dir[key].filename
        files.append(
            get_coverage_data(
                root_info,
                covdata_dir[key],
                cdata_sourcefile[fname],
                cdata_fname[fname],
                relative_path,
            )
        )

    directory_data["entries"] = files

    return directory_data


def get_file_data(
    options: Options,
    root_info: RootInfo,
    filename: str,
    cdata_fname: dict[str, str],
    cdata_sourcefile: dict[str, str],
    cdata: FileCoverage,
) -> tuple[dict[str, Any], dict[tuple[str, str, int], dict[str, Any]], bool]:
    """Get the data for a file to generate the HTML"""
    formatter = get_formatter(options)

    file_data = dict[str, Any](
        {
            "fname": filename,
            "filename": cdata_fname[filename],
            "html_filename": os.path.basename(cdata_sourcefile[filename]),
            "source_lines": [],
            "function_list": [],
        }
    )
    file_data.update(
        get_coverage_data(
            root_info, cdata, cdata_sourcefile[filename], cdata_fname[filename]
        )
    )
    functions = dict[tuple[FunctioncovKeyType, str, int], dict[str, Any]]()
    # Only use demangled names (containing a brace)
    for functioncov in cdata.functioncov(key=lambda functioncov: functioncov.key):
        for lineno in functioncov.linenos:
            f_data = dict[str, Any]()
            f_data["name"] = functioncov.name
            f_data["filename"] = cdata_fname[filename]
            f_data["html_filename"] = os.path.basename(cdata_sourcefile[filename])
            f_data["line"] = lineno
            f_data["count"] = functioncov.count[lineno]
            f_data["blocks"] = functioncov.blocks[lineno]
            f_data["excluded"] = functioncov.excluded[lineno]
            function_stats = cdata.filter_for_function(functioncov).stats
            f_data["line_coverage"] = function_stats.line.percent_or(100.0)
            f_data["branch_coverage"] = function_stats.branch.percent_or("-")
            f_data["condition_coverage"] = function_stats.condition.percent_or("-")

            file_data["function_list"].append(f_data)
            functions[
                (
                    functioncov.key,
                    str(f_data["filename"]),
                    int(f_data["line"]),
                )
            ] = f_data

    def get_linecovs(lineno: int) -> list[LineCoverage] | None:
        """Get a list of line coverage objects if available for the line."""
        linecovs = cdata.get_line(lineno)
        return None if linecovs is None else list(linecovs.linecov(sort=True))

    with chdir(options.root_dir):
        linecov_collections = list(cdata.lines())
        max_line_from_cdata = (
            linecov_collections[-1].lineno if linecov_collections else 0
        )
        file_not_found = True
        try:
            with open(
                filename,
                "r",
                encoding=options.source_encoding,
                errors="replace",
            ) as source_file:
                file_not_found = False
                lines = formatter.highlighter_for_file(filename)(source_file.read())
                lineno = 0
                for lineno, line in enumerate(lines, 1):
                    file_data["source_lines"].append(
                        source_row(
                            lineno,
                            line,
                            get_linecovs(lineno),
                            options.html_block_ids,
                        )
                    )
                if lineno < max_line_from_cdata:
                    LOGGER.warning(
                        "File %s has %d line(s) but coverage data has %d line(s).",
                        filename,
                        lineno,
                        max_line_from_cdata,
                    )
        except OSError as e:
            if filename.endswith("<stdin>"):
                file_not_found = False
                file_info = "!!! File from stdin !!!"
            else:
                file_info = f"!!! Can't read file: {e.strerror} !!!"
                LOGGER.warning("Can't read file: %s", e)
            # Python ranges are exclusive. We want to iterate over all lines, including
            # that last line. Thus, we have to add a +1 to include that line.
            for lineno in range(1, max_line_from_cdata + 1):
                file_data["source_lines"].append(
                    source_row(
                        lineno,
                        file_info if lineno == 1 else "",
                        get_linecovs(lineno),
                        options.html_block_ids,
                    )
                )

    file_data["lines"]["uncovered"] = len(
        [row for row in file_data["source_lines"] if row["covclass"] == "uncoveredLine"]
    )
    file_data["lines"]["partial"] = len(
        [
            row
            for row in file_data["source_lines"]
            if row["covclass"] == "partialCoveredLine"
        ]
    )

    return file_data, functions, file_not_found


def dict_from_stat(
    stat: CoverageStat | DecisionCoverageStat,
    coverage_class: Callable[[float | None], str],
    default: float | None = None,
) -> dict[str, Any]:
    """Get a dictionary from the stats."""
    coverage_default = "-" if default is None else default
    data = {
        "total": stat.total_with_excluded
        if isinstance(stat, CoverageStat)
        else stat.total,
        "exec": stat.covered,
        "excluded": stat.excluded if isinstance(stat, CoverageStat) else "-",
        "coverage": stat.percent_or(coverage_default),
        "class": coverage_class(stat.percent_or(default)),
    }

    if isinstance(stat, DecisionCoverageStat):
        data["unchecked"] = stat.uncheckable

    return data


def source_row(
    lineno: int,
    source: str,
    linecov_list: list[LineCoverage] | None,
    html_block_ids: bool,
) -> dict[str, Any]:
    """Get information for a row"""
    line_branches = None
    line_conditions = None
    line_decisions = None
    line_calls = None
    linecount: int | str = ""
    covclass = ""
    if linecov_list:
        if all(linecov.is_excluded for linecov in linecov_list):
            covclass = "excludedLine"
        elif all(linecov.is_uncovered for linecov in linecov_list):
            covclass = "uncoveredLine"
        else:
            line_branches = [
                source_row_branch(linecov)
                for linecov in linecov_list
                if linecov.has_reportable_branches
            ]
            covclass = (
                "coveredLine"
                if all(
                    line_branch["taken"] == line_branch["total"]
                    for line_branch in line_branches
                )
                else "partialCoveredLine"
            )
            line_conditions = [
                source_row_condition(linecov)
                for linecov in linecov_list
                if linecov.has_reportable_conditions
            ]
            line_decisions = [
                source_row_decision(linecov)
                for linecov in linecov_list
                if linecov.decision
            ]
            linecount = sum(linecov.count for linecov in linecov_list)
        line_calls = [
            source_row_call(linecov)
            for linecov in linecov_list
            if linecov.has_reportable_calls
        ]
    return {
        "lineno": lineno,
        "block_ids": []
        if linecov_list is None
        or not html_block_ids
        or all(linecov is None for linecov in linecov_list)
        else [linecov.block_ids for linecov in linecov_list if linecov.block_ids],
        "source": source,
        "covclass": covclass,
        "line_branches": line_branches,
        "line_conditions": line_conditions,
        "line_decisions": line_decisions,
        "line_calls": line_calls,
        "linecount": linecount,
    }


def source_row_branch(
    linecov: LineCoverage,
) -> dict[str, Any]:
    """Get branch information for a row"""
    items = list[dict[str, Any]]()
    for branchcov in [
        branchcov
        for branchcov in linecov.branches(sort=True)
        if branchcov.is_reportable
    ]:
        items.append(
            {
                "taken": branchcov.is_covered,
                "count": branchcov.count,
                "excluded": branchcov.is_excluded,
            }
        )
        if branchcov.branchno is not None:
            items[-1]["branchno"] = branchcov.branchno
        else:
            items[-1]["source_block_id"] = branchcov.source_block_id
            items[-1]["destination_block_id"] = branchcov.destination_block_id

    stats = linecov.branch_coverage()
    return {
        "function_name": linecov.report_function_name,
        "taken": stats.covered,
        "total": stats.total_with_excluded,
        "branches": items,
    }


def source_row_condition(
    linecov: LineCoverage,
) -> dict[str, Any]:
    """Get condition information for a row."""

    items = []

    conditioncov_list = list(
        conditioncov
        for conditioncov in linecov.conditions(sort=True)
        if conditioncov.is_reportable
    )
    for conditioncov in conditioncov_list:
        condition_prefix = (
            f"Condition {conditioncov.conditionno}"
            if len(conditioncov_list) > 1
            else ("" if conditioncov.count == 2 else "Condition ")
        )
        condition_separator = "." if len(conditioncov_list) > 1 else ""
        for index in range(0, conditioncov.count // 2):
            items.append(
                {
                    "prefix": f"{condition_prefix}{condition_separator}{index}: "
                    if conditioncov.count > 2
                    else (f"{condition_prefix}: " if condition_prefix else ""),
                    "not_covered_true": index in conditioncov.not_covered_true,
                    "not_covered_false": index in conditioncov.not_covered_false,
                    "excluded": conditioncov.is_excluded,
                }
            )
    stats = linecov.condition_coverage()
    return {
        "function_name": linecov.report_function_name,
        "count": stats.total_with_excluded,
        "covered": stats.covered,
        "condition": items,
    }


def source_row_decision(
    linecov: LineCoverage,
) -> dict[str, Any]:
    """Get decision information for a row"""

    items: list[dict[str, Any]] = []

    if isinstance(linecov.decision, DecisionCoverageUncheckable):
        items.append(
            {
                "uncheckable": True,
            }
        )
    elif isinstance(linecov.decision, DecisionCoverageConditional):
        items.append(
            {
                "uncheckable": False,
                "taken": linecov.decision.count_true > 0,
                "count": linecov.decision.count_true,
                "name": "true",
            }
        )
        items.append(
            {
                "uncheckable": False,
                "taken": linecov.decision.count_false > 0,
                "count": linecov.decision.count_false,
                "name": "false",
            }
        )
    elif isinstance(linecov.decision, DecisionCoverageSwitch):
        items.append(
            {
                "uncheckable": False,
                "taken": linecov.decision.count > 0,
                "count": linecov.decision.count,
                "name": "true",
            }
        )
    else:
        raise RuntimeError(f"Unknown decision type {linecov.decision!r}")

    return {
        "function_name": linecov.report_function_name,
        "taken": len([i for i in items if i.get("taken", False)]),
        "uncheckable": len([i for i in items if i["uncheckable"]]),
        "total": len(items),
        "decisions": items,
    }


def source_row_call(linecov: LineCoverage) -> dict[str, Any]:
    """Get call information for a source row."""
    items = []

    for callno, callcov in enumerate(
        callcov for callcov in linecov.calls(sort=True) if callcov.is_reportable
    ):
        items.append(
            {
                "name": callno,
                "invoked": callcov.is_covered,
                "excluded": callcov.is_excluded,
            }
        )

    stats = linecov.call_coverage()
    return {
        "function_name": linecov.report_function_name,
        "invoked": stats.covered,
        "total": stats.total_with_excluded,
        "calls": items,
    }


def _get_prefix_and_suffix(output_file: str) -> tuple[str, str]:
    """Split into prefix and suffix, ignoring the last suffix for GZIP."""
    (output_prefix, output_suffix) = os.path.splitext(os.path.abspath(output_file))
    if output_suffix == GZIP_SUFFIX:
        return _get_prefix_and_suffix(output_prefix)

    return (output_prefix, output_suffix)


def _make_short_source_filename(output_file: str, filename: str) -> str:
    r"""Make a short-ish file path for --html-detail output.

    Args:
        output_file (str): The output path.
        filename (str): Path from root to source code.
    """

    (output_prefix, output_suffix) = _get_prefix_and_suffix(output_file)
    filename = filename.replace(os.sep, "/").replace("<stdin>", "stdin")
    source_filename = (
        ".".join(
            (
                output_prefix,
                os.path.basename(filename),
                get_md5_hexdigest(filename.encode("UTF-8")),
            )
        )
        + output_suffix
    )
    return source_filename



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

import gzip
import json
import os
from glob import glob

from ...filter import is_file_excluded

from ...data_model import version
from ...data_model.container import CoverageContainer
from ...data_model.merging import get_merge_mode_from_options
from ...logging import LOGGER
from ...options import Options
from ...utils import GZIP_SUFFIX


#
#  Get coverage from already existing gcovr JSON files
#
def read_report(options: Options) -> CoverageContainer:
    """Read trace files into internal data model."""

    covdata = CoverageContainer()
    if len(options.json_tracefile) != 0:
        datafiles = set()

        for trace_file_pattern in options.json_tracefile:
            trace_files = glob(trace_file_pattern, recursive=True)
            if not trace_files:
                raise RuntimeError(
                    f"Bad --json-add-tracefile={trace_file_pattern} option.\n"
                    "\tThe specified file does not exist."
                )

            for activate_trace_logging in trace_files:
                datafiles.add(os.path.normpath(activate_trace_logging))

        merge_options = get_merge_mode_from_options(options)
        for data_source in datafiles:
            activate_trace_logging = not is_file_excluded(
                "trace",
                data_source,
                options.trace_include_filter,
                options.trace_exclude_filter,
            )
            if activate_trace_logging:
                LOGGER.trace("Processing file: %s", data_source)

            if data_source.casefold().endswith(GZIP_SUFFIX):
                with gzip.open(data_source, "rt", encoding="utf-8") as fh:
                    gcovr_json_data = json.loads(fh.read())
            else:
                with open(data_source, encoding="utf-8") as json_file:
                    gcovr_json_data = json.load(json_file)

            format_version = str(gcovr_json_data["gcovr/format_version"])
            if format_version != version.FORMAT_VERSION:
                raise AssertionError(
                    f"Wrong format version, got {format_version} expected {version.FORMAT_VERSION}."
                )

            covdata.merge(
                CoverageContainer.deserialize(
                    data_source, gcovr_json_data["files"], options, merge_options
                ),
                merge_options,
            )

    return covdata



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

"""
Handle parsing of the textual ``.gcov`` file format.

Other modules should only use the following items:
`parse_metadata()`, `parse_coverage()`

The behavior of this parser was informed by the following sources:

* the old GcovParser class
  <https://github.com/gcovr/gcovr/blob/e0b7afef00123b7b6ce4f487a1c4cc9fc60528bc/gcovr/gcov.py#L239>
* the *Invoking Gcov* section in the GCC manual (version 11)
  <https://gcc.gnu.org/onlinedocs/gcc-11.1.0/gcc/Invoking-Gcov.html>
* the ``gcov.c`` source code in GCC
  (especially for understanding the exact number format)
  <https://github.com/gcc-mirror/gcc/blob/releases/gcc-11.1.0/gcc/gcov.c>
"""
# pylint: disable=too-many-lines

import enum
import re
from typing import Any, Iterable, NamedTuple, Pattern

from ....data_model.coverage import FileCoverage, LineCoverage
from ....data_model.merging import FUNCTION_MAX_LINE_MERGE_OPTIONS, MergeOptions
from ....exceptions import SanityCheckError
from ....logging import LOGGER
from ....utils import get_md5_hexdigest

from .common import (
    SUSPICIOUS_COUNTER,
    check_hits,
)

UNKNOWN_FUNCTION_NAME_FORMAT_STRING = "<unknown function {}>"


def _line_pattern(pattern: str) -> Pattern[str]:
    """
    Compile a regex from a line pattern.

    A line pattern is a normal regex, except that the following placeholders
    will be replaced by pattern fragments:

    * ``VALUE`` -> matches gcov's ``format_gcov()`` output (percentage or
      human-readable)
    * ``INT`` -> matches an integer
    * the pattern is anchored at the start/end
    * space is replaced by ``[ ]+``
    """
    pattern = pattern.replace(" ", r" +")
    pattern = pattern.replace("INT", r"[0-9]+")
    pattern = pattern.replace("VALUE", r"(?:NAN %|-?[0-9.]+[%kMGTPEZY]?)")
    return re.compile("^" + pattern + "$")


_RE_FUNCTION_LINE = _line_pattern(
    r"function (.*?) called (INT) returned (VALUE) blocks executed (VALUE)"
)
_RE_BRANCH_LINE = _line_pattern(
    r"branch (INT) (?:taken (VALUE)|never executed)(?: \((\w+)\))?"
)
_RE_CALL_LINE = _line_pattern(r"call (INT) (?:returned (VALUE)|never executed)")
_RE_UNCONDITIONAL_LINE = _line_pattern(
    r"unconditional (INT) (?:taken (VALUE)|never executed)"
)
_RE_SOURCE_LINE = _line_pattern(r"(?: )?(VALUE[*]?|-|[#]{5}|[=]{5}):(?: )?(INT):(.*)")
_RE_BLOCK_LINE = _line_pattern(r"(?: )?(VALUE|[$]{5}|[%]{5}):(?: )?(INT)-block (INT)")


class _ExtraInfo(enum.Flag):
    """Additional info about lines, such as noncode or exception-only status."""

    NONE = 0
    NONCODE = enum.auto()
    EXCEPTION_ONLY = enum.auto()
    PARTIAL = enum.auto()

    def __repr__(self) -> str:
        return str(self).replace("_ExtraInfo.", "")


class _SourceLine(NamedTuple):
    """A gcov line with source code: ``HITS: LINENO:CODE``"""

    hits: int
    lineno: int
    source_code: str
    extra_info: _ExtraInfo


class _MetadataLine(NamedTuple):
    """A gcov line with metadata: ``-: 0:KEY:VALUE``"""

    key: str
    value: str | None


class _BlockLine(NamedTuple):
    """A gcov line with block data: ``HITS: LINENO-block BLOCKNO``"""

    hits: int
    lineno: int
    block_id: int
    extra_info: _ExtraInfo


class _CallLine(NamedTuple):
    """A gcov line with call data: ``call CALLNO returned RETURNED``"""

    callno: int
    returned: int


class _BranchLine(NamedTuple):
    """A gcov line with branch data: ``branch BRANCHNO taken HITS (ANNOTATION)``"""

    branchno: int
    hits: int
    annotation: str | None


class _UnconditionalLine(NamedTuple):
    """
    A gcov line with unconditional branch data: ``unconditional BRANCHNO taken HITS``
    """

    branchno: int
    hits: int


class _FunctionLine(NamedTuple):
    """
    A gcov line with function coverage data for the next line.

    ``function NAME called COUNT returned RETURNED blocks executed BLOCKS``
    """

    name: str
    call_count: int | None
    blocks_covered: float | None


class _FunctionSpecializationNameLine(NamedTuple):
    """A gcov line with the name of a specialization section: ``NAME:``"""

    name: str


class _FunctionSpecializationSeparatorLine(NamedTuple):
    """A gcov line that delimits function specializations (no fields)"""


# NamedTuples can't inherit from a common base,
# so we use a Union type as the _parse_line() return type.
#
# Why NamedTuples? Better type safety than tuples, but very low memory overhead.
_Line = (
    _SourceLine
    | _MetadataLine
    | _BlockLine
    | _CallLine
    | _BranchLine
    | _UnconditionalLine
    | _FunctionLine
    | _FunctionSpecializationNameLine
    | _FunctionSpecializationSeparatorLine
)


class UnknownLineType(Exception):
    """Used by `_parse_line()` to signal that no known line type matched."""

    def __init__(self, line: str) -> None:
        super().__init__(line)
        self.line = line


def parse_metadata(
    filename: str,
    lines: list[str],
    *,
    suspicious_hits_threshold: int = SUSPICIOUS_COUNTER,
    activate_trace_logging: bool = False,
) -> dict[str, str | None]:
    r"""
    Collect the header/metadata lines from a gcov file.

    Example:
    >>> parse_metadata("file", '''
    ...   -: 0:Foo:bar
    ...   -: 0:Key:123
    ... '''.splitlines())
    Traceback (most recent call last):
       ...
    RuntimeError: Missing key 'Source' in metadata. GCOV data was >>
      -: 0:Foo:bar
      -: 0:Key:123<< End of GCOV data
    >>> parse_metadata("file", '-: 0:Source: file \n -: 0:Foo: bar \n -: 0:Key: 123 '.splitlines())
    {'Source': 'file', 'Foo': 'bar', 'Key': '123'}
    >>> parse_metadata("file", '''
    ...   -: 0:Source:file
    ...   -: 0:Foo:bar
    ...   -: 0:Key
    ... '''.splitlines())
    {'Source': 'file', 'Foo': 'bar', 'Key': None}
    """
    collected = {}
    for line in lines:
        # empty lines shouldn't occur in reality, but are common in testing
        if not line:
            continue

        parsed_line = _parse_line(filename, line, suspicious_hits_threshold)

        if isinstance(parsed_line, _MetadataLine):
            key, value = parsed_line
            if activate_trace_logging:
                LOGGER.trace("Metadata: %s=%s", parsed_line.key, parsed_line.value)
            collected[key] = value
        else:
            break  # stop at the first line that is not metadata

    if "Source" not in collected:
        data = "\n".join(lines)
        raise RuntimeError(
            f"Missing key 'Source' in metadata. GCOV data was >>{data}<< End of GCOV data"
        )

    return collected


_LineWithError = tuple[str, Exception]


def parse_coverage(
    data_filename: str | set[tuple[str, ...]],
    lines: list[str],
    *,
    filename: str,
    ignore_parse_errors: set[str] | None,
    suspicious_hits_threshold: int = SUSPICIOUS_COUNTER,
    activate_trace_logging: bool = False,
    use_existing_files: bool = False,
) -> tuple[FileCoverage, list[str]]:
    """
    Extract coverage data from a gcov report.

    Logging:
    Parse problems are reported as warnings.
    Coverage exclusion decisions are reported as verbose messages.

    Arguments:
        lines: the lines of the file to be parsed (excluding newlines)
        filename: for error reports
        ignore_parse_errors: which errors should be converted to warnings

    Returns:
        tuple of the coverage data and the source code lines

    Raises:
        Any exceptions during parsing, unless ignore_parse_errors is set.
    """

    lines_with_errors = list[_LineWithError]()
    tokenized_lines = list[tuple[_Line, str]]()
    persistent_states = dict[str, Any]()
    for raw_line in lines:
        # empty lines shouldn't occur in reality, but are common in testing
        if not raw_line:
            continue

        try:
            parsed_line = _parse_line(
                filename,
                raw_line,
                suspicious_hits_threshold,
                ignore_parse_errors,
                persistent_states,
            )
            tokenized_lines.append((parsed_line, raw_line))
        except Exception as ex:  # pylint: disable=broad-except
            lines_with_errors.append((raw_line, ex))

    missing_function_lines = not any(
        isinstance(line, _FunctionLine) for line, _ in tokenized_lines
    )

    if (
        use_existing_files
        # Only check for missing function lines if we have any non-metadata lines.
        and any(not isinstance(line, _MetadataLine) for line, _ in tokenized_lines)
        and missing_function_lines
    ):
        files_for_message = (
            "\n   ".join(
                " -> ".join(file_tuple) for file_tuple in sorted(data_filename)
            )
            if isinstance(data_filename, set)
            else data_filename
        )
        LOGGER.warning(
            f"No function line found in gcov file:\n   {files_for_message}\n"
            "This may indicate that the file was generated without "
            "the proper gcov options (especially --branch-probabilities)?\n"
            "See <https://gcovr.com/en/stable/faq.html#which-options-are-used-for-calling-gcov>.",
        )

    if (
        "negative_hits.warn_once_per_file" in persistent_states
        and persistent_states["negative_hits.warn_once_per_file"] > 1
    ):
        LOGGER.warning(
            "Ignored %d negative hits overall.",
            persistent_states["negative_hits.warn_once_per_file"],
        )

    if (
        "suspicious_hits.warn_once_per_file" in persistent_states
        and persistent_states["suspicious_hits.warn_once_per_file"] > 1
    ):
        LOGGER.warning(
            "Ignored %d suspicious hits overall.",
            persistent_states["suspicious_hits.warn_once_per_file"],
        )

    filecov = FileCoverage(data_filename, filename=filename)
    state = _ParserState.new(
        function_name=UNKNOWN_FUNCTION_NAME_FORMAT_STRING,
    )
    for line, raw_line in tokenized_lines:
        try:
            if activate_trace_logging:
                LOGGER.trace("Processing line: %s", line)
            state = _gather_coverage_from_line(
                state,
                line,
                filecov=filecov,
                activate_trace_logging=activate_trace_logging,
            )
        except Exception as ex:  # pylint: disable=broad-except
            lines_with_errors.append((raw_line, ex))
            state = _ParserState.new(is_recovering=True)

    # Clean up the final state. This shouldn't happen,
    # but the last line could theoretically contain pending function lines
    for function in state.deferred_functions:
        name, count, blocks = function
        filecov.insert_function_coverage(
            str(data_filename),
            MergeOptions(func_opts=FUNCTION_MAX_LINE_MERGE_OPTIONS),
            mangled_name=name,
            demangled_name=None,
            lineno=(state.last_linecov.lineno + 1) if state.last_linecov else 0,
            count=count,
            blocks=blocks,
        )

    if state.unknown_function is not None:
        filecov.insert_function_coverage(
            str(data_filename),
            MergeOptions(func_opts=FUNCTION_MAX_LINE_MERGE_OPTIONS),
            mangled_name=state.unknown_function[0],
            demangled_name=None,
            lineno=state.unknown_function[1],
            count=None,
            blocks=None,
            excluded=True,
        )

    _report_lines_with_errors(
        lines_with_errors,
        filename=filename,
        ignore_parse_errors=ignore_parse_errors,
    )

    src_lines = _reconstruct_source_code(line for line, _ in tokenized_lines)

    return filecov, src_lines


def _reconstruct_source_code(tokens: Iterable[_Line]) -> list[str]:
    source_token_lines = [line for line in tokens if isinstance(line, _SourceLine)]

    src_lines = [""] * max((line.lineno for line in source_token_lines), default=0)
    for line in source_token_lines:
        src_lines[line.lineno - 1] = line.source_code

    return src_lines


class _ParserState(NamedTuple):
    """State information while parsing gcov lines.

    >>> state = _ParserState.new()
    >>> state.previous_state is None
    True
    >>> state = state._replace(linecov_list=["x"])
    >>> state.linecov_list == ["x"]
    True
    >>> new_state = state.save_state()
    >>> new_state == state
    False
    >>> new_state.previous_state == state
    True
    >>> new_state.linecov_list == ["x"]
    True
    >>> state.linecov_list == ["x"]
    True
    >>> state = new_state._replace(linecov_list=[])
    >>> state.linecov_list == []
    True
    >>> restored_state = state.restore_state()
    >>> restored_state.previous_state is None
    True
    >>> state.linecov_list == []
    True
    >>> restored_state.restore_state()
    Traceback (most recent call last):
    ...
    gcovr.exceptions.SanityCheckError: Previous_state of _ParserState is None.
    """

    deferred_functions: list[_FunctionLine] = []
    function_name: str | None = None
    function_specialization: bool = False
    last_linecov: LineCoverage | None = None
    linecov_list: list[LineCoverage] = []
    block_id: int | None = None
    is_recovering: bool = False
    previous_state: "_ParserState | None" = None
    unknown_function: tuple[str, int] | None = None

    @classmethod
    def new(cls, **kwargs: Any) -> "_ParserState":
        """Return a newly initialized parser state object."""
        # We need to overwrite the list values because the default values are shared
        # and changed.
        return (
            cls()
            ._replace(
                deferred_functions=[],
                linecov_list=[],
            )
            ._replace(**kwargs)
        )

    def save_state(self) -> "_ParserState":
        """Save the current parser state and return a new one with the current line coverage list."""
        return _ParserState.new(
            previous_state=self,
            linecov_list=self.linecov_list,
            unknown_function=self.unknown_function,
        )

    def restore_state(self) -> "_ParserState":
        """Restore the previous parser state with the current line coverage list."""
        if self.previous_state is None:
            raise SanityCheckError(f"Previous_state of {type(self).__name__} is None.")
        return self.previous_state._replace(
            linecov_list=self.linecov_list,
            unknown_function=self.unknown_function,
        )


def _gather_coverage_from_line(
    state: _ParserState,
    line: _Line,
    *,
    filecov: FileCoverage,
    activate_trace_logging: bool,
) -> _ParserState:
    """
    Interpret a Line, updating the FileCoverage, and transitioning ParserState.

    The function handles all possible Line variants, and dies otherwise:
    >>> _gather_coverage_from_line(_ParserState(), "illegal line type", filecov=..., activate_trace_logging=...)
    Traceback (most recent call last):
    AssertionError: Unexpected line type: 'illegal line type'
    """
    # pylint: disable=too-many-return-statements,too-many-branches
    # pylint: disable=no-else-return  # make life easier for type checkers
    if isinstance(line, _SourceLine):
        raw_count, lineno, source_code, extra_info = line

        # handle deferred functions
        if state.deferred_functions:
            for function in state.deferred_functions:
                name, count, blocks = function
                filecov.insert_function_coverage(
                    filecov.data_sources,
                    MergeOptions(func_opts=FUNCTION_MAX_LINE_MERGE_OPTIONS),
                    mangled_name=name,
                    demangled_name=None,
                    lineno=lineno,
                    count=count,
                    blocks=blocks,
                )
            state.deferred_functions.clear()

        # If a specialized function is following a normal one the overall lines were added
        # already to the previous function and we need to remove them again.
        if (
            state.function_specialization
            and state.linecov_list
            and state.linecov_list[0].lineno <= lineno <= state.linecov_list[-1].lineno
        ):
            to_remove = [
                linecov for linecov in state.linecov_list if linecov.lineno >= lineno
            ]

            for linecov in to_remove:
                if activate_trace_logging:
                    LOGGER.trace(
                        "%s: Removing line coverage of function %s.",
                        linecov.location,
                        linecov.report_function_name,
                    )
                filecov.remove_line_coverage(linecov)

            if state.unknown_function == (
                to_remove[0].function_name,
                to_remove[0].lineno,
            ):
                state = state._replace(unknown_function=None)

            state.linecov_list.clear()

        if not extra_info & _ExtraInfo.NONCODE:
            # Add line number to unknown function name and insert the function coverage object.
            if state.function_name == UNKNOWN_FUNCTION_NAME_FORMAT_STRING:
                unknown_function_name = UNKNOWN_FUNCTION_NAME_FORMAT_STRING.format(
                    lineno
                )
                state = state._replace(
                    function_name=unknown_function_name,
                    unknown_function=(unknown_function_name, lineno),
                )

            linecov = filecov.insert_line_coverage(
                filecov.data_sources,
                lineno=lineno,
                count=raw_count,
                function_name=state.function_name,
                md5=get_md5_hexdigest(source_code.encode("UTF-8")),
            )
            state = state._replace(last_linecov=linecov)
            if not state.function_specialization:
                state.linecov_list.append(linecov)

        return state

    elif state.is_recovering:
        return state  # skip until the next _SourceLine

    elif isinstance(line, _FunctionLine):
        # Defer handling of the function tag until the next source line.
        # This is important to get correct line number information.
        if state.deferred_functions and activate_trace_logging:
            LOGGER.trace(
                "Several function definitions for the next coverage lines, function %s will not contain line coverage.",
                state.deferred_functions[-1].name,
            )
        return state._replace(
            deferred_functions=[*state.deferred_functions, line],
            function_name=line.name,
        )

    elif isinstance(line, _FunctionSpecializationNameLine):
        # Now we know that we are in a specialization and not at the end of it.
        return state.save_state()._replace(
            function_specialization=True,
        )

    elif isinstance(line, _FunctionSpecializationSeparatorLine):
        # If there was a specialization active we need to restore the previous state, else we do nothing.
        if state.function_specialization:
            state = state.restore_state()

        return state

    elif isinstance(line, _BranchLine):
        branchno, hits, annotation = line

        if state.last_linecov is not None:
            state.last_linecov.insert_branch_coverage(
                filecov.data_sources,
                branchno=branchno,
                count=hits,
                source_block_id=state.block_id,
                fallthrough=(annotation == "fallthrough"),
                throw=(annotation == "throw"),
            )

        return state

    # ignore unused line types, such as specialization sections
    elif isinstance(line, _CallLine):
        callno, returned = line

        # linecov won't exist if it was considered noncode
        if state.last_linecov is not None:
            state.last_linecov.insert_call_coverage(
                filecov.data_sources,
                callno=callno,
                source_block_id=state.block_id,  # type: ignore [arg-type]
                destination_block_id=None,
                returned=returned,
            )

        return state

    elif isinstance(line, _BlockLine):
        return state._replace(block_id=line.block_id)

    # ignore metadata in this phase
    elif isinstance(line, _MetadataLine):
        return state

    elif isinstance(line, (_UnconditionalLine,)):
        return state

    raise AssertionError(f"Unexpected line type: {line!r}")


def _report_lines_with_errors(
    lines_with_errors: list[_LineWithError],
    *,
    filename: str,
    ignore_parse_errors: set[str] | None,
) -> None:
    """Log warnings and potentially re-throw exceptions"""

    if not lines_with_errors:
        return

    lines = [line for line, _ in lines_with_errors]
    errors = [error for _, error in lines_with_errors]

    lines_output = "\n\t  ".join(lines)
    LOGGER.warning(
        "Unrecognized GCOV output for %s\n\t  %s\n"
        "\tThis is indicative of a gcov output parse error.\n"
        "\tPlease report this to the gcovr developers\n"
        "\tat <https://github.com/gcovr/gcovr/issues>.",
        filename,
        lines_output,
    )

    for ex in errors:
        LOGGER.warning("Exception during parsing:\n\t%s: %s", type(ex).__name__, ex)

    if ignore_parse_errors is not None and "all" in ignore_parse_errors:
        return

    LOGGER.error(
        "Exiting because of parse errors.\n"
        "\tYou can run gcovr with --gcov-ignore-parse-errors=...\n"
        "\tto continue anyway."
    )

    # if we caught an exception, re-raise it for the traceback
    raise errors[0]  # guaranteed to have at least one exception


def _parse_line(
    filename: str,
    line: str,
    suspicious_hits_threshold: int = SUSPICIOUS_COUNTER,
    ignore_parse_errors: set[str] | None = None,
    persistent_states: dict[str, Any] | None = None,
) -> _Line:
    """
    Categorize/parse individual lines without further processing.

    Example: can parse code line:
    >>> _parse_line("file", '     -: 13:struct Foo{};')
    _SourceLine(hits=0, lineno=13, source_code='struct Foo{};', extra_info=NONCODE)
    >>> _parse_line("file", '    12: 13:foo += 1;  ')
    _SourceLine(hits=12, lineno=13, source_code='foo += 1;  ', extra_info=NONE)
    >>> _parse_line("file", ' #####: 13:foo += 1;')
    _SourceLine(hits=0, lineno=13, source_code='foo += 1;', extra_info=NONE)
    >>> _parse_line("file", ' #####:10000:foo += 1;')  # see https://github.com/gcovr/gcovr/issues/882
    _SourceLine(hits=0, lineno=10000, source_code='foo += 1;', extra_info=NONE)
    >>> _parse_line("file", ' =====: 13:foo += 1;')
    _SourceLine(hits=0, lineno=13, source_code='foo += 1;', extra_info=EXCEPTION_ONLY)
    >>> _parse_line("file", '   12*: 13:cond ? f() : g();')
    _SourceLine(hits=12, lineno=13, source_code='cond ? f() : g();', extra_info=PARTIAL)
    >>> _parse_line("file", ' 1.7k*: 13:foo();')
    _SourceLine(hits=1700, lineno=13, source_code='foo();', extra_info=PARTIAL)

    Example: can parse metadata line:
    >>> _parse_line("file", '  -: 0:Foo:bar baz')
    _MetadataLine(key='Foo', value='bar baz')
    >>> _parse_line("file", '  -: 0:Some key:2')  # coerce numbers
    _MetadataLine(key='Some key', value='2')

    Example: can parse branch tags:
    >>> _parse_line("file", 'branch 3 taken 15%')
    _BranchLine(branchno=3, hits=1, annotation=None)
    >>> _parse_line("file", 'branch 3 taken 0%')
    _BranchLine(branchno=3, hits=0, annotation=None)
    >>> _parse_line("file", 'branch 3 taken 123')
    _BranchLine(branchno=3, hits=123, annotation=None)
    >>> _parse_line("file", 'branch 3 taken -1', ignore_parse_errors=("negative_hits.warn",))
    _BranchLine(branchno=3, hits=0, annotation=None)
    >>> _parse_line("file", 'branch 3 taken 4294967296', ignore_parse_errors=("suspicious_hits.warn",))
    _BranchLine(branchno=3, hits=0, annotation=None)
    >>> _parse_line("file", 'branch 7 taken 3% (fallthrough)')
    _BranchLine(branchno=7, hits=1, annotation='fallthrough')
    >>> _parse_line("file", 'branch 17 taken 99% (throw)')
    _BranchLine(branchno=17, hits=1, annotation='throw')
    >>> _parse_line("file", 'branch  0 never executed')
    _BranchLine(branchno=0, hits=0, annotation=None)
    >>> _parse_line("file", 'branch  0 never executed (fallthrough)')
    _BranchLine(branchno=0, hits=0, annotation='fallthrough')
    >>> _parse_line("file", 'branch 2 with some unknown format')
    Traceback (most recent call last):
    gcovr.formats.gcov.parser.text.UnknownLineType: branch 2 with some unknown format

    Example: can parse call tags:
    >>> _parse_line("file", 'call  0 never executed')
    _CallLine(callno=0, returned=0)
    >>> _parse_line("file", 'call  17 returned 50%')
    _CallLine(callno=17, returned=1)
    >>> _parse_line("file", 'call  17 returned 9')
    _CallLine(callno=17, returned=9)
    >>> _parse_line("file", 'call 2 with some unknown format')
    Traceback (most recent call last):
    gcovr.formats.gcov.parser.text.UnknownLineType: call 2 with some unknown format

    Example: can parse unconditional branches
    >>> _parse_line("file", 'unconditional 1 taken 17')
    _UnconditionalLine(branchno=1, hits=17)
    >>> _parse_line("file", 'unconditional 2 taken -1', ignore_parse_errors=set(['negative_hits.warn']))
    _UnconditionalLine(branchno=2, hits=0)
    >>> _parse_line("file", 'unconditional 2 taken 4294967296', ignore_parse_errors=set(['suspicious_hits.warn']))
    _UnconditionalLine(branchno=2, hits=0)
    >>> _parse_line("file", 'unconditional 3 never executed')
    _UnconditionalLine(branchno=3, hits=0)
    >>> _parse_line("file", 'unconditional with some unknown format')
    Traceback (most recent call last):
    gcovr.formats.gcov.parser.text.UnknownLineType: unconditional with some unknown format

    Example: can parse function tags:
    >>> _parse_line("file", 'function foo called 2 returned 1 blocks executed 85%')
    _FunctionLine(name='foo', call_count=2, blocks_covered=85.0)
    >>> _parse_line("file", 'function foo called 2 returned 50% blocks executed 85%')
    _FunctionLine(name='foo', call_count=2, blocks_covered=85.0)
    >>> _parse_line("file", 'function foo called 2 returned 100% blocks executed 85%')
    _FunctionLine(name='foo', call_count=2, blocks_covered=85.0)
    >>> _parse_line("file", 'function foo with some unknown format')
    Traceback (most recent call last):
    gcovr.formats.gcov.parser.text.UnknownLineType: function foo with some unknown format

    Example: can parse template specialization markers:
    >>> _parse_line("file", '------------------')
    _FunctionSpecializationSeparatorLine()

    Example: can parse template specialization names:
    >>> _parse_line("file", 'Foo<bar>::baz():')
    _FunctionSpecializationNameLine(name='Foo<bar>::baz()')
    >>> _parse_line("file", ' foo:')
    Traceback (most recent call last):
    gcovr.formats.gcov.parser.text.UnknownLineType:  foo:
    >>> _parse_line("file", ':')
    Traceback (most recent call last):
    gcovr.formats.gcov.parser.text.UnknownLineType: :

    Example: can parse block line:
    >>> _parse_line("file", '     1: 32-block  0')
    _BlockLine(hits=1, lineno=32, block_id=0, extra_info=NONE)
    >>> _parse_line("file", ' %%%%%: 33-block  1')
    _BlockLine(hits=0, lineno=33, block_id=1, extra_info=NONE)
    >>> _parse_line("file", ' $$$$$: 33-block  1')
    _BlockLine(hits=0, lineno=33, block_id=1, extra_info=EXCEPTION_ONLY)
    >>> _parse_line("file", ' %%%%%:10000-block  0')  # see https://github.com/gcovr/gcovr/issues/882
    _BlockLine(hits=0, lineno=10000, block_id=0, extra_info=NONE)
    >>> _parse_line("file", '     -1: 32-block  0', ignore_parse_errors=set(['negative_hits.warn']))
    _BlockLine(hits=0, lineno=32, block_id=0, extra_info=NONE)
    >>> _parse_line("file", '     4294967296: 32-block  0', ignore_parse_errors=set(['suspicious_hits.warn']))
    _BlockLine(hits=0, lineno=32, block_id=0, extra_info=NONE)
    >>> _parse_line("file", '     1: 9-block with some unknown format')
    Traceback (most recent call last):
    gcovr.formats.gcov.parser.text.UnknownLineType:      1: 9-block with some unknown format

    Example: will reject garbage:
    >>> _parse_line("file", 'nonexistent_tag foo bar')
    Traceback (most recent call last):
    gcovr.formats.gcov.parser.text.UnknownLineType: nonexistent_tag foo bar
    """
    # pylint: disable=too-many-branches
    if ignore_parse_errors is None:
        ignore_parse_errors = set()
    if persistent_states is None:
        persistent_states = {"location": (filename, 0)}

    tag = _parse_tag_line(
        line,
        suspicious_hits_threshold,
        ignore_parse_errors,
        persistent_states,
    )
    if tag is not None:
        return tag

    # Handle lines that are like source lines.
    # But this could also include metadata lines and block-coverage lines.

    # CODE
    #
    # Structure: "COUNT: LINENO:CODE"
    #
    # Examples:
    #     -: 13:struct Foo{};
    #    12: 13:foo += 1;
    # #####: 13:foo += 1;
    # =====: 13:foo += 1;
    #   12*: 13:cond ? bar() : baz();
    match = _RE_SOURCE_LINE.fullmatch(line)
    if match is not None:
        hits_str, lineno, source_code = match.groups()
        persistent_states.update(location=(filename, int(lineno)))

        # METADATA (key, value)
        if hits_str == "-" and lineno == "0":
            if ":" in source_code:
                key, value = source_code.split(":", 1)
                return _MetadataLine(key, value.strip())

            # Add a synthetic metadata with no value
            return _MetadataLine(source_code, None)

        if hits_str == "-":
            hits = 0
            extra_info = _ExtraInfo.NONCODE
        elif hits_str == "#####":
            hits = 0
            extra_info = _ExtraInfo.NONE
        elif hits_str == "=====":
            hits = 0
            extra_info = _ExtraInfo.EXCEPTION_ONLY
        elif hits_str.endswith("*"):
            hits = _int_from_gcov_unit(hits_str[:-1])
            extra_info = _ExtraInfo.PARTIAL
        else:
            hits = _int_from_gcov_unit(hits_str)
            extra_info = _ExtraInfo.NONE

        hits = check_hits(
            hits,
            line,
            ignore_parse_errors,
            suspicious_hits_threshold,
            persistent_states,
        )

        return _SourceLine(hits, int(lineno), source_code, extra_info)

    # BLOCK
    #
    # Structure: "COUNT: LINENO-block BLOCKNO"
    if "-block " in line:
        match = _RE_BLOCK_LINE.match(line)
        if match is not None:
            hits_str, lineno, block_id = match.groups()
            persistent_states.update(location=(filename, int(lineno)))

            if hits_str == "%%%%%":
                hits = 0
                extra_info = _ExtraInfo.NONE
            elif hits_str == "$$$$$":
                hits = 0
                extra_info = _ExtraInfo.EXCEPTION_ONLY
            else:
                hits = _int_from_gcov_unit(hits_str)
                extra_info = _ExtraInfo.NONE

            hits = check_hits(
                hits,
                line,
                ignore_parse_errors,
                suspicious_hits_threshold,
                persistent_states,
            )

            return _BlockLine(hits, int(lineno), int(block_id), extra_info)

    # SPECIALIZATION NAME
    #
    # Structure: a name starting in the first column, ending with a ":". It is
    # not safe to make further assumptions about the layout of the (demangled)
    # identifier. For example, Rust might produce "<X as Y>::foo::h12345".
    #
    # This line type is therefore checked LAST! The old parser might have been
    # more robust because it would only consider specialization names on the
    # line following a specialization marker.
    if len(line) > 2 and not line[0].isspace() and line.endswith(":"):
        return _FunctionSpecializationNameLine(line[:-1])

    raise UnknownLineType(line)


def _parse_tag_line(  # pylint: disable=too-many-return-statements
    line: str,
    suspicious_hits_threshold: int,
    ignore_parse_errors: set[str],
    persistent_states: dict[str, Any],
) -> _Line | None:
    """A tag line is any gcov line that starts in the first column."""

    # Tag lines never start with whitespace.
    #
    # In principle, specialization names are also like tag lines.
    # But they don't have a marker, so their detection is done last.
    if line.startswith(" "):
        return None

    # BRANCH
    #
    # Structure:
    # branch BRANCHNO never executed
    # branch BRANCHNO taken VALUE
    # branch BRANCHNO taken VALUE (ANNOTATION)
    if line.startswith("branch "):
        match = _RE_BRANCH_LINE.match(line)
        if match is not None:
            branch_id, taken_str, annotation = match.groups()
            hits = 0 if taken_str is None else _int_from_gcov_unit(taken_str)

            hits = check_hits(
                hits,
                line,
                ignore_parse_errors,
                suspicious_hits_threshold,
                persistent_states,
            )

            return _BranchLine(int(branch_id), hits, annotation)

    # CALL
    #
    # Structure (note whitespace after tag):
    # call  0 never executed
    # call  1 returned VALUE
    if line.startswith("call "):
        match = _RE_CALL_LINE.match(line)
        if match is not None:
            call_id, returned_str = match.groups()
            returned = 0 if returned_str is None else _int_from_gcov_unit(returned_str)
            return _CallLine(int(call_id), returned)

    # UNCONDITIONAL
    #
    # Structure:
    # unconditional NUM taken VALUE
    # unconditional NUM never executed
    if line.startswith("unconditional "):
        match = _RE_UNCONDITIONAL_LINE.match(line)
        if match is not None:
            branch_id, taken_str = match.groups()
            hits = 0 if taken_str is None else _int_from_gcov_unit(taken_str)

            hits = check_hits(
                hits,
                line,
                ignore_parse_errors,
                suspicious_hits_threshold,
                persistent_states,
            )

            return _UnconditionalLine(int(branch_id), hits)

    # FUNCTION
    #
    # Structure:
    # function NAME called VALUE returned VALUE blocks executed VALUE
    if line.startswith("function "):
        match = _RE_FUNCTION_LINE.match(line)
        if match is not None:
            name, count, _, blocks = match.groups()
            return _FunctionLine(
                name, _int_from_gcov_unit(count), _float_from_gcov_percent(blocks)
            )

    # Function separator line,
    # see https://github.com/gcc-mirror/gcc/blob/50bc9185c2821350f0b785d6e23a6e9dcde58466/gcc/gcov.c#L3100C23-L3100C41
    #
    # Structure: literally just lots of hyphens
    if line == "------------------":
        return _FunctionSpecializationSeparatorLine()

    return None


def _int_from_gcov_unit(formatted: str) -> int:
    """
    Try to reverse gcov's number formatting.

    Gcov's number formatting works like this:

    * if ``decimal_places >= 0``, format a percentage
      * the percentage is fudged so that 0% and 100% are only shown
        when that's the true value
    * otherwise, format a count
      * if human readable numbers are enabled,
        use SI units like ``1.7k`` instead of ``1693``

    Relevant gcov command line flags:

    * ``-c`` enables counts instead of percentages
    * ``-H`` enables human-readable numbers (SI units)

    Note that percentages destroy information: the original value can't be recovered,
    so we must map to zero/one.
    Of course, counts are not that useful either because we don't know the max value.

    Examples:
    >>> _int_from_gcov_unit('123')
    123
    >>> _int_from_gcov_unit('-1.2k')
    -1200
    >>> [_int_from_gcov_unit(value) for value in ('NAN %', '17.2%', '0%')]
    [0, 1, 0]
    >>> [_int_from_gcov_unit(value) for value in ('1.7k', '0.5G')]
    [1700, 500000000]
    """
    if formatted.endswith("%"):
        return 1 if float(formatted[:-1]) > 0 else 0

    units = "kMGTPEZY"
    for exponent, unit in enumerate(units, 1):
        if formatted.endswith(unit):
            return int(float(formatted[:-1]) * 1000**exponent)

    return int(formatted)


def _float_from_gcov_percent(formatted: str) -> float:
    """
    Transform percentage to float value

    Examples:
    >>> [_float_from_gcov_percent(value) for value in ('NAN %', '17.2%', '0%')]
    [nan, 17.2, 0.0]
    """

    if not formatted.endswith("%"):
        raise AssertionError(f"Number must end with %, got {formatted}")

    return float(formatted[:-1])



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

import os
from glob import glob
from lxml import etree  # nosec # We only write XML files

from ...data_model.container import CoverageContainer
from ...data_model.coverage import (
    BranchCoverage,
    FileCoverage,
    LineCoverage,
)
from ...data_model.merging import MergeOptions, get_merge_mode_from_options
from ...filter import is_file_excluded
from ...logging import LOGGER
from ...options import Options


#
#  Get coverage from already existing gcovr JSON files
#
def read_report(options: Options) -> CoverageContainer:
    """merge a coverage from multiple reports in the format
    compatible with Cobertura"""

    covdata = CoverageContainer()
    if len(options.cobertura_tracefile) == 0:
        return covdata

    datafiles = set()

    for trace_files_regex in options.cobertura_tracefile:
        trace_files = glob(trace_files_regex, recursive=True)
        if not trace_files:
            raise RuntimeError(
                "Bad --covertura-add-tracefile option.\n"
                "\tThe specified file does not exist."
            )

        for activate_trace_logging in trace_files:
            datafiles.add(os.path.normpath(activate_trace_logging))

    for data_sources in datafiles:
        LOGGER.debug("Processing XML file: %s", data_sources)

        try:
            root: etree._Element = etree.parse(data_sources).getroot()  # nosec # We parse the file given by the user
        except Exception as e:
            raise RuntimeError(f"Bad --cobertura-add-tracefile option.\n{e}") from None

        source_elem = root.find("./sources/source")
        if source_elem is None:
            raise AssertionError(f"No source directory defined in file {data_sources}")
        source_dir = str(source_elem.text)

        gcovr_file: etree._Element
        for gcovr_file in root.xpath("./packages//class"):  # type: ignore [assignment, union-attr]
            filename = gcovr_file.get("filename")
            if filename is None:  # pragma: no cover
                LOGGER.warning(
                    "Missing filename attribute in class element at %s:%s",
                    data_sources,
                    gcovr_file.sourceline,
                )
                continue

            filename = str(os.path.normpath(os.path.join(source_dir, filename)))
            if is_file_excluded(
                "source file", filename, options.include_filter, options.exclude_filter
            ):
                continue

            filecov = FileCoverage(data_sources, filename=filename)
            merge_options = get_merge_mode_from_options(options)
            xml_line: etree._Element
            for xml_line in gcovr_file.xpath("./lines//line"):  # type: ignore [assignment, union-attr]
                _insert_line_from_xml(filecov, data_sources, merge_options, xml_line)

            covdata.insert_file_coverage(filecov, merge_options)

    return covdata


def _insert_line_from_xml(
    filecov: FileCoverage,
    data_sources: str,
    merge_options: MergeOptions,
    xml_line: etree._Element,
) -> None:
    try:
        lineno = int(xml_line.get("number", ""))
    except Exception:  # pragma: no cover
        raise RuntimeError(
            "Bad --covertura-add-tracefile option.\n"
            f"'number' attribute is required and must be an integer: {etree.tostring(xml_line).decode()}\n"
        ) from None

    try:
        count = int(xml_line.get("hits", ""))
    except Exception:  # pragma: no cover
        raise RuntimeError(
            "Bad --covertura-add-tracefile option.\n"
            f"'hits' attribute is required and must be an integer: {etree.tostring(xml_line).decode()}\n"
        ) from None

    is_branch = xml_line.get("branch") == "true"
    branch_msg = xml_line.get("condition-coverage")
    linecov = filecov.insert_line_coverage(
        data_sources, merge_options, lineno=lineno, count=count, function_name=None
    )

    if is_branch and branch_msg is not None:
        try:
            [covered, total] = branch_msg[branch_msg.rfind("(") + 1 : -1].split("/")
            for i in range(int(total)):
                _branch_from_json(linecov, data_sources, i, i < int(covered))
        except AssertionError as exc:  # pragma: no cover
            LOGGER.warning(
                "Invalid branch information for line %s: %s", linecov.lineno, exc
            )


def _branch_from_json(
    linecov: LineCoverage, data_sources: str, branchno: int, is_covered: bool
) -> BranchCoverage:
    return linecov.insert_branch_coverage(
        data_sources,
        branchno=branchno,
        count=1 if is_covered else 0,
    )



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

import gzip
from json import loads as json_loads, dumps as json_dumps
import os
import re
import shlex
import subprocess  # nosec # Commands are trusted.
from threading import Lock
from typing import Any, Callable

from ...data_model.container import CoverageContainer
from ...data_model.merging import get_merge_mode_from_options
from ...exceptions import SanityCheckError
from ...decision_analysis import DecisionParser
from ...exclusions import (
    apply_all_exclusions,
    get_exclusion_options_from_options,
)
from ...filter import Filter, is_file_excluded
from ...logging import LOGGER
from ...options import Options
from ...utils import (
    PRETTY_JSON_INDENT,
    commonpath,
    fix_case_of_path,
    is_fs_case_insensitive,
    search_file,
)
from .parser import (
    json,
    text,
)
from .workers import Workers, locked_directory

output_re = re.compile(r"[Cc]reating [`'](.*)'$")
source_error_re = re.compile(
    r"(?:[Cc](?:annot|ould not) open (?:source|graph|notes) file|: No such file or directory)"
)
output_error_re = re.compile(
    r"(?:[Cc](?:annot|ould not) open output file|Operation not permitted|Permission denied|Read-only file system)"
)
version_mismatch_re = re.compile(r":version '[^']+', prefer.*'[^']+'")


def read_report(options: Options) -> CoverageContainer:
    """Read data from GCOV output."""
    datafiles = set()

    find_files = find_datafiles
    process_file = process_datafile
    if options.gcov_use_existing_files:
        find_files = find_existing_gcov_files
        process_file = process_existing_gcov_file

    # Get data files
    if not options.search_paths:
        options.search_paths = [options.root]

        if options.gcov_objdir is not None:
            options.search_paths.append(options.gcov_objdir)

    for search_path in options.search_paths:
        datafiles.update(find_files(search_path, options.exclude_directory))

    # Get coverage data
    with Workers(
        options.gcov_parallel,
        lambda: {"covdata": CoverageContainer(), "to_erase": set(), "options": options},
    ) as pool:
        LOGGER.debug("Pool started with %d threads", pool.size())
        for filename in sorted(datafiles):
            pool.add(process_file, filename)
        try:
            contexts = pool.wait()
        except KeyboardInterrupt as exc:
            # Stop the pool if Ctrl+C is pressed
            pool.drain()
            raise exc from None

    to_erase = set()
    covdata = CoverageContainer()
    for context in contexts:
        covdata.merge(context["covdata"], get_merge_mode_from_options(options))
        to_erase.update(context["to_erase"])

    for filepath in to_erase:
        if os.path.exists(filepath):
            os.remove(filepath)

    return covdata


def find_existing_gcov_files(
    search_path: str, exclude_directory: list[re.Pattern[str]]
) -> list[str]:
    """Find .gcov and .gcov.json.gz files under the given search path."""
    if os.path.isfile(search_path):
        LOGGER.debug("Using given gcov file %s", search_path)
        gcov_files = [search_path]
    else:
        LOGGER.debug("Scanning directory %s for gcov files...", search_path)
        gcov_files = list(
            search_file(
                lambda fname: re.compile(r".*\.gcov(?:\.json\.gz)?$").match(fname)
                is not None,
                search_path,
                exclude_directory=exclude_directory,
            )
        )
        LOGGER.debug("Found %d files (and will process all of them)", len(gcov_files))
    return gcov_files


def find_datafiles(
    search_path: str, exclude_directory: list[re.Pattern[str]]
) -> list[str]:
    """Find .gcda and .gcno files under the given search path.

    The .gcno files will *only* produce uncovered results.
    However, that is useful information when a compilation unit
    is never actually exercised by the test code.
    So we ONLY return them if there's no corresponding .gcda file.
    """
    if os.path.isfile(search_path):
        LOGGER.debug(
            "Using given %s file %s",
            os.path.splitext(search_path)[1][1:],
            search_path,
        )
        files = [search_path]
    else:
        LOGGER.debug("Scanning directory %s for gcda/gcno files...", search_path)
        files = list(
            search_file(
                lambda fname: re.compile(r".*\.gc(da|no)$").match(fname) is not None,
                search_path,
                exclude_directory=exclude_directory,
            )
        )
    gcda_files = []
    gcno_files = []
    known_file_stems = set()
    for filename in files:
        stem, ext = os.path.splitext(filename)
        if ext == ".gcda":
            gcda_files.append(filename)
            known_file_stems.add(stem)
        elif ext == ".gcno":
            gcno_files.append(filename)
    # remove gcno files that match a gcno stem
    gcno_files = [
        filename
        for filename in gcno_files
        if os.path.splitext(filename)[0] not in known_file_stems
    ]
    LOGGER.debug(
        "Found %d files (and will process %d)",
        len(files),
        len(gcda_files) + len(gcno_files),
    )
    return gcda_files + gcno_files


#
# Process a single gcov datafile
#
def process_gcov_json_data(
    data_fname: str,
    covdata: CoverageContainer,
    options: Options,
) -> None:
    """Process a GCOV JSON output."""
    activate_trace_logging = not is_file_excluded(
        "trace", data_fname, options.trace_include_filter, options.trace_exclude_filter
    )

    with gzip.open(data_fname, "rt", encoding="utf-8") as fh_in:
        gcov_json_data = json_loads(fh_in.read())
        if activate_trace_logging:
            LOGGER.trace(
                "Parsing gcov data file %s:\n%s<<EOF",
                data_fname,
                json_dumps(gcov_json_data, indent=PRETTY_JSON_INDENT),
            )

    merge_options = get_merge_mode_from_options(options)
    for filecov, source_lines in json.parse_coverage(
        data_fname,
        gcov_json_data,
        include_filter=options.include_filter,
        exclude_filter=options.exclude_filter,
        ignore_parse_errors=options.gcov_ignore_parse_errors,
        suspicious_hits_threshold=options.gcov_suspicious_hits_threshold,
        source_encoding=options.source_encoding,
        activate_trace_logging=activate_trace_logging,
    ):
        activate_trace_logging = not is_file_excluded(
            "trace",
            filecov.filename,
            options.trace_include_filter,
            options.trace_exclude_filter,
        )
        if activate_trace_logging:
            LOGGER.trace("Apply exclusions for %s", filecov.filename)
        apply_all_exclusions(
            filecov,
            lines=source_lines,
            options=get_exclusion_options_from_options(options),
            activate_trace_logging=activate_trace_logging,
        )

        if options.show_decision:
            decision_parser = DecisionParser(filecov, source_lines)
            decision_parser.parse_all_lines()

        if activate_trace_logging:
            LOGGER.trace(
                "Merge coverage data for %s using %s.", filecov.filename, merge_options
            )
        covdata.insert_file_coverage(filecov, merge_options)


#
# Process a single gcov datafile
#
def process_gcov_text_data(
    data_fname: str,
    gcda_fname: str | None,
    covdata: CoverageContainer,
    options: Options,
    current_dir: str | None = None,
) -> None:
    """Process a GCOV text output."""
    activate_trace_logging = not is_file_excluded(
        "trace", data_fname, options.trace_include_filter, options.trace_exclude_filter
    )
    with open(
        data_fname, "r", encoding=options.source_encoding, errors="replace"
    ) as fh_in:
        content = fh_in.read()
        if activate_trace_logging:
            LOGGER.trace("Parsing gcov data file %s:\n%s<<EOF", data_fname, content)
        lines = content.splitlines()

    # Find the source file
    metadata = text.parse_metadata(
        data_fname,
        lines,
        suspicious_hits_threshold=options.gcov_suspicious_hits_threshold,
        activate_trace_logging=activate_trace_logging,
    )
    source = metadata.get("Source")
    if source is None:
        raise RuntimeError("Unexpected value 'None' for metadata 'Source'.")
    # gcov writes filenames with '/' path separators even if the OS
    # separator is different, so we replace it with the correct separator
    source = source.replace("/", os.sep)

    fname = guess_source_file_name(
        source,
        data_fname,
        gcda_fname,
        root_dir=options.root_dir,
        starting_dir=options.starting_dir,
        obj_dir=(
            None
            if options.gcov_objdir is None
            else os.path.abspath(options.gcov_objdir)
        ),
        current_dir=current_dir,
    )

    if is_file_excluded(
        "source file", fname, options.include_filter, options.exclude_filter
    ):
        return

    if activate_trace_logging:
        LOGGER.trace("Parsing coverage data for file %s", fname)
    key = os.path.normpath(fname)

    filecov, source_lines = text.parse_coverage(
        set([(gcda_fname, data_fname) if gcda_fname else (data_fname,)]),
        lines,
        filename=key,
        ignore_parse_errors=options.gcov_ignore_parse_errors,
        suspicious_hits_threshold=options.gcov_suspicious_hits_threshold,
        activate_trace_logging=activate_trace_logging,
        use_existing_files=options.gcov_use_existing_files,
    )

    if activate_trace_logging:
        LOGGER.trace("Apply exclusions for %s", fname)
    apply_all_exclusions(
        filecov,
        lines=source_lines,
        options=get_exclusion_options_from_options(options),
        activate_trace_logging=activate_trace_logging,
    )

    if options.show_decision:
        decision_parser = DecisionParser(filecov, source_lines)
        decision_parser.parse_all_lines()

    merge_mode = get_merge_mode_from_options(options)
    if activate_trace_logging:
        LOGGER.trace("Merge coverage data for %s using %s.", fname, merge_mode)
    covdata.insert_file_coverage(filecov, merge_mode)


def guess_source_file_name(
    source_from_gcov: str,
    data_fname: str,
    gcda_fname: str | None,
    root_dir: str,
    starting_dir: str,
    obj_dir: str | None,
    current_dir: str | None = None,
) -> str:
    """Guess the full source filename."""
    if current_dir is None:
        current_dir = os.getcwd()
    if os.path.isabs(source_from_gcov):
        fname = source_from_gcov
    elif gcda_fname is None:
        fname = guess_source_file_name_via_aliases(
            source_from_gcov, data_fname, current_dir
        )
    else:
        fname = guess_source_file_name_heuristics(
            source_from_gcov,
            data_fname,
            gcda_fname,
            current_dir,
            root_dir,
            starting_dir,
            obj_dir,
        )

    if is_fs_case_insensitive():
        fname = fix_case_of_path(fname)

    LOGGER.debug(
        "Finding source file corresponding to a gcov data file\n"
        "  gcov_fname   %s\n"
        "  current_dir  %s\n"
        "  root         %s\n"
        "  starting_dir %s\n"
        "  obj_dir      %s\n"
        "  gcda_fname   %s\n"
        "  --> fname    %s",
        data_fname,
        current_dir,
        root_dir,
        starting_dir,
        obj_dir,
        gcda_fname,
        fname,
    )

    return fname


def guess_source_file_name_via_aliases(
    source_from_gcov: str,
    data_fname: str,
    current_dir: str,
) -> str:
    """Guess the full source filename with path by an alias."""
    common_dir = commonpath([data_fname, current_dir])
    fname = os.path.abspath(os.path.join(common_dir, source_from_gcov))
    if os.path.exists(fname):
        return fname

    initial_fname = fname

    data_fname_dir = os.path.dirname(data_fname)
    fname = os.path.abspath(os.path.join(data_fname_dir, source_from_gcov))
    if os.path.exists(fname):
        return fname

    # @latk-2018: The original code is *very* insistent
    # on returning the initial guess. Why?
    return initial_fname


def guess_source_file_name_heuristics(  # pylint: disable=too-many-return-statements
    source_from_gcov: str,
    data_fname: str,
    gcda_fname: str,
    current_dir: str,
    root_dir: str,
    starting_dir: str,
    obj_dir: str | None,
) -> str:
    """Guess the full source filename with path by a heuristic."""
    # 0. Try using the path to the gcov file
    fname = os.path.join(os.path.dirname(data_fname), source_from_gcov)
    if os.path.exists(fname):
        return fname

    LOGGER.debug("Fallback to heuristic of gcovr 5.1")

    # 1. Try using the current working directory as the source directory
    fname = os.path.join(current_dir, source_from_gcov)
    if os.path.exists(fname):
        return fname

    # 2. Try using the path to common prefix with the root_dir as the source directory
    fname = os.path.join(root_dir, source_from_gcov)
    if os.path.exists(fname):
        return fname

    # 3. Try using the starting directory as the source directory
    fname = os.path.join(starting_dir, source_from_gcov)
    if os.path.exists(fname):
        return fname

    # 4. Try using relative path from object dir
    if obj_dir is not None:
        fname = os.path.normpath(os.path.join(obj_dir, source_from_gcov))
        if os.path.exists(fname):
            return fname

    # Get path of gcda file
    gcda_fname_dir = os.path.dirname(gcda_fname)

    # 5. Try using the path to the gcda as the source directory
    fname = os.path.join(gcda_fname_dir, source_from_gcov)
    if os.path.exists(fname):
        return os.path.normpath(fname)

    # 6. Try using the path to the gcda file as the source directory, removing the path part from the gcov file
    fname = os.path.join(gcda_fname_dir, os.path.basename(source_from_gcov))
    return fname


def process_datafile(
    filename: str, covdata: CoverageContainer, options: Options, to_erase: set[str]
) -> None:
    r"""Run gcovr in a suitable directory to collect coverage from gcda files.

    Params:
        filename (path): the path to a gcda or gcno file
        covdata (dict, mutable): the global covdata dictionary
        options (object): the configuration options namespace
        to_erase (set, mutable): files that should be deleted later

    Returns:
        Nothing.

    Finding a suitable working directory is tricky.
    The coverage files (gcda and gcno) are stored next to object (.o) files.
    However, gcov needs to also resolve the source file name.
    The relative source file paths in the coverage data
    are relative to the gcc working directory.
    Therefore, gcov must be invoked in the same directory as gcc.
    How to find that directory? By various heuristics.

    This is complicated by the problem that the build process tells gcc
    where to run, where the sources are, and where to put the object files.
    We only know the object files and have to work everything out in reverse.

    Ideally, the build process only runs gcc from *one* directory
    and the user can provide this directory as the ``--gcov-object-directory``.
    If it exists, we try that path as a work dir,
    If the path is relative, it is resolved relative to the gcovr cwd and the
    object file location.

    We next try the ``--root`` directory.
    TODO: should probably also be the gcovr start directory.

    If none of those work, we assume that
    the object files are in a subdirectory of the gcc working directory,
    i.e. we can walk the directory tree upwards.

    All of this works fine unless gcc was invoked like ``gcc -o ../path``,
    i.e. the object files are in a sibling directory.
    TODO: So far there is no good way to address this case.
    """
    activate_trace_logging = not is_file_excluded(
        "trace", filename, options.trace_include_filter, options.trace_exclude_filter
    )
    if activate_trace_logging:
        LOGGER.trace("Processing file: %s", filename)

    abs_filename = os.path.abspath(filename).replace(
        os.path.sep, "/"
    )  # gcov requires posix style path

    errors = list[str]()

    potential_wd = []

    if options.gcov_objdir:
        potential_wd = find_potential_working_directories_via_objdir(
            abs_filename, options.gcov_objdir, error=errors.append
        )

    # no objdir was specified or objdir didn't exist
    consider_parent_directories = not potential_wd

    # Always add the root directory
    potential_wd.append(options.root_dir)

    if consider_parent_directories:
        wd = os.path.dirname(abs_filename)
        while wd != potential_wd[-1]:
            potential_wd.append(wd)
            wd = os.path.dirname(wd)

    for wd in potential_wd:
        done = run_gcov_and_process_files(
            abs_filename,
            covdata,
            options=options,
            error=errors.append,
            chdir=wd,
        )

        if options.delete_input_files:
            if not abs_filename.endswith("gcno"):
                to_erase.add(abs_filename)

        if done:
            return

    # Join the errors with proper indention
    errors_output = "\n\t".join("\n\t\t".join(e.split("\n")) for e in errors)
    errors_output = (
        f"GCOV produced the following errors processing {abs_filename!r}:\n"
        f"\t{errors_output}\n"
        "GCOVR could not infer a working directory that resolved it.\n"
        f"{'' if options.verbose else 'Use option --verbose to get extended information. '}"
        "To ignore this error use option --gcov-ignore-errors=no_working_dir_found."
    )

    # Check if error shall be ignored
    if options.gcov_ignore_errors is None or not any(
        v in options.gcov_ignore_errors for v in ["all", "no_working_dir_found"]
    ):
        raise RuntimeError(errors_output)


def find_potential_working_directories_via_objdir(
    abs_filename: str, objdir: str, error: Callable[[str], None]
) -> list[str]:
    """Find the potential working directories."""
    # absolute path - just return the objdir
    if os.path.isabs(objdir):
        if os.path.isdir(objdir):
            return [objdir]

    # relative path: check relative to both the cwd and the gcda file
    else:
        potential_wd = [
            testdir
            for prefix in [os.path.dirname(abs_filename), os.getcwd()]
            for testdir in [os.path.join(prefix, objdir)]
            if os.path.isdir(testdir)
        ]

        if potential_wd:
            return potential_wd

    error(
        "ERROR: cannot identify the location where GCC "
        f"was run using --gcov-object-directory={objdir}\n"
    )

    return []


class GcovProgram:
    """Class to execute GCOV command with a set of auto-detected options"""

    __lock = Lock()
    __cmd: str = ""
    __cmd_split = list[str]()
    __default_options = list[str]()
    __exitcode_to_ignore = list[int]([0])
    __help_output: str = ""
    __version_output: str = ""

    class LockContext:
        """Context handler for locking a section in multithreaded executions."""

        def __init__(self, lock: Lock) -> None:
            self.lock = lock

        def __enter__(self) -> None:
            self.lock.acquire()

        def __exit__(self, *_: Any) -> None:
            self.lock.release()

    class GcovExecutionError(Exception):
        """Exception for errors in gcov execution."""

    def __init__(self, cmd: str) -> None:
        with GcovProgram.LockContext(GcovProgram.__lock):
            if not GcovProgram.__cmd:
                GcovProgram.__cmd = cmd
                # If the first element of cmd - the executable name - has embedded spaces
                # (other than within quotes), it probably includes extra arguments.
                GcovProgram.__cmd_split = shlex.split(GcovProgram.__cmd)
            elif GcovProgram.__cmd != cmd:
                raise AssertionError(
                    f"Gcov command must not be changed, expected '{GcovProgram.__cmd}', got '{cmd}'"
                )

    @classmethod
    def reset(cls) -> None:
        """Reset the cached values (for testing purposes)."""
        with GcovProgram.LockContext(GcovProgram.__lock):
            cls.__cmd = ""
            cls.__cmd_split = list[str]()
            cls.__default_options = list[str]()
            cls.__exitcode_to_ignore = list[int]([0])
            cls.__help_output = ""
            cls.__version_output = ""

    def identify_and_cache_capabilities(self) -> None:
        """Check the capabilities of GCOVR once."""
        with GcovProgram.LockContext(GcovProgram.__lock):
            if not GcovProgram.__default_options:
                GcovProgram.__default_options = [
                    "--branch-counts",
                    "--branch-probabilities",
                    "--all-blocks",
                ]

                if self.__check_gcov_help_content("--json-format"):
                    if self.__check_gcov_version_content(
                        f"JSON format version: {json.GCOV_JSON_VERSION}"
                    ):
                        LOGGER.debug("GCOV capabilities: JSON format available.")
                        GcovProgram.__default_options.append("--json-format")
                        if self.__check_gcov_help_content("--conditions"):
                            LOGGER.debug(
                                "GCOV capabilities: Condition coverage available."
                            )
                            GcovProgram.__default_options.append("--conditions")
                    else:
                        LOGGER.debug(
                            "GCOV capabilities: Unsupported JSON format detected."
                        )

                if self.__check_gcov_help_content("--demangled-names"):
                    LOGGER.debug("GCOV capabilities: Demangled names available.")
                    GcovProgram.__default_options.append("--demangled-names")

                if self.__check_gcov_help_content("--hash-filenames"):
                    LOGGER.debug("GCOV capabilities: Hashing of filenames available.")
                    GcovProgram.__default_options.append("--hash-filenames")
                elif self.__check_gcov_help_content("--preserve-paths"):
                    LOGGER.debug("GCOV capabilities: Preserve of paths available.")
                    GcovProgram.__default_options.append("--preserve-paths")
                else:
                    LOGGER.warning(
                        "Options '--hash-filenames' and '--preserve-paths' are not supported by '%s'. Source files with identical file names may result in incorrect coverage.",
                        GcovProgram.__cmd,
                    )

                if not self.__check_gcov_help_content("LLVM"):
                    GcovProgram.__exitcode_to_ignore.append(6)  # WRITE GCOV ERROR

    def __get_help_output(self) -> str:
        if not GcovProgram.__help_output:
            GcovProgram.__help_output = ""
            for help_option in ["--help", "--help-hidden"]:
                gcov_process = self.__get_gcov_process(
                    [help_option],
                    universal_newlines=True,
                )
                out, _ = gcov_process.communicate(timeout=30)

                if not gcov_process.returncode:
                    # gcov execution was successful, help argument is not supported.
                    GcovProgram.__help_output += out
            if not GcovProgram.__help_output:
                # gcov tossed errors: throw exception
                raise RuntimeError("Error in gcov command line, couldn't get help.")

        return GcovProgram.__help_output

    def __get_version_output(self) -> str:
        if not GcovProgram.__version_output:
            gcov_process = self.__get_gcov_process(
                ["--version"],
                universal_newlines=True,
            )
            out, _ = gcov_process.communicate(timeout=30)

            if gcov_process.returncode:  # pragma: no cover
                # gcov tossed errors: throw exception
                raise RuntimeError(
                    "Error in gcov command line, couldn't get version information."
                )
            # gcov execution was successful, help argument is not supported.
            GcovProgram.__version_output = out

        return GcovProgram.__version_output

    def __check_gcov_help_content(self, option: str) -> bool:
        if option in self.__get_help_output():
            return True

        return False

    def __check_gcov_version_content(self, option: str) -> bool:
        if option in self.__get_version_output():
            return True

        return False

    def get_default_options(self) -> list[str]:
        """Get the default options for GCOV."""
        return GcovProgram.__default_options

    def __get_gcov_process(
        self, args: list[str], trace: bool = False, **kwargs: Any
    ) -> "subprocess.Popen[str]":
        # NB: Currently, we will only parse English output
        env = kwargs.pop("env") if "env" in kwargs else dict(os.environ)
        env["LC_ALL"] = "C"
        env["LANGUAGE"] = "en_US"

        if "cwd" not in kwargs:
            kwargs["cwd"] = "."
        cmd = GcovProgram.__cmd_split + args
        if trace:
            LOGGER.trace("Running gcov in %s: %s", kwargs["cwd"], shlex.join(cmd))

        return subprocess.Popen(  # nosec # We know that we execute gcov tool
            cmd,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
            **kwargs,
        )

    def run_with_args(
        self,
        args: list[str],
        cwd: str,
        activate_trace_logging: bool = False,
        **kwargs: Any,
    ) -> tuple[str, str]:
        """Run the gcov program.

        >>> import platform
        >>> if platform.system() == "Windows":
        ...     print("kill not working on Windows")  # doctest: +SKIP
        ... else:
        ...     GcovProgram("bash").run_with_args(["-c", "exit 1"])
        Traceback (most recent call last):
        ...
        GcovProgram.GcovExecutionError: GCOV returncode was 1.
        >>> if platform.system() == "Windows":
        ...     GcovProgram("bash").run_with_args(["-c", "exit 1"])
        ... else:
        ...     print("kill not working on Windows")  # doctest: +SKIP
        Traceback (most recent call last):
        ...
        GcovProgram.GcovExecutionError: GCOV returncode was 4294967295.
        >>> if platform.system() == "Windows":
        ...     print("kill not working on Windows")  # doctest: +SKIP
        ... else:
        ...     GcovProgram("bash").run_with_args(["-c", "kill $$"])
        Traceback (most recent call last):
        ...
        GcovProgram.GcovExecutionError: GCOV returncode was -15 (exited by signal).
        >>> if platform.system() == "Windows":
        ...     GcovProgram("bash").run_with_args(["-c", "kill $$"])
        ... else:
        ...     print("kill not working on Windows")  # doctest: +SKIP
        Traceback (most recent call last):
        ...
        GcovProgram.GcovExecutionError: GCOV returncode was 15.
        """
        process = self.__get_gcov_process(
            args, cwd=cwd, trace=activate_trace_logging, **kwargs
        )
        out, err = process.communicate()

        def remove_generated_files() -> None:
            """Remove the generated files from gcov output."""
            for line in out.splitlines():
                found = output_re.search(line.strip())
                if found is not None:
                    fname = found.group(1)
                    if not os.path.isabs(fname):
                        fname = os.path.join(cwd, fname)
                    if os.path.exists(fname):
                        os.remove(fname)

        if (
            process.returncode < 0
            or process.returncode not in GcovProgram.__exitcode_to_ignore
        ):
            remove_generated_files()
            raise self.GcovExecutionError(
                f"GCOV returncode was {process.returncode}{' (exited by signal)' if process.returncode < 0 else ''}.\n"
                f"STDERR >>{err}<< End of STDERR\n"
                f"STDOUT >>{out}<< End of STDOUT"
            )

        if version_mismatch_re.search(err):
            # gcov tossed errors: throw exception
            raise self.GcovExecutionError(
                f"Version mismatch gcc/gcov.\nSTDERR >>{err}<< End of STDERR"
            )

        if activate_trace_logging:
            LOGGER.trace("STDERR >>%s<< End of STDERR", err)
            LOGGER.trace("STDOUT >>%s<< End of STDOUT", out)

        return (out, err)


def run_gcov_and_process_files(
    abs_filename: str,
    covdata: CoverageContainer,
    options: Options,
    error: Callable[[str], None],
    chdir: str,
) -> bool:
    """Run GCOV tool and process the output files."""

    done = False

    # ATTENTION:
    # This lock is essential for parallel processing because without
    # this there can be name collisions for the generated output files.
    with locked_directory(chdir):

        def remove_existing_files(files: list[str]) -> None:
            """Remove the existing files from the given list."""
            for filepath in sorted(files):
                if os.path.exists(filepath):
                    os.remove(filepath)

        class GcovMessageOnStderr(Exception):
            """Exception for errors messages of gcov printed to STDOUT."""

        filename = None
        out = None
        err = None
        active_gcov_files = set[str]()
        try:
            gcov_cmd = GcovProgram(options.gcov_cmd)
            gcov_cmd.identify_and_cache_capabilities()

            filename = abs_filename
            # Use try catch because the relpath can fail on Windows for different drives.
            # Do not know how to force this exception therefore ignore coverage.
            try:
                filename = os.path.relpath(filename, chdir)
            except OSError:  # pragma: no cover # nosec
                pass
            object_directory = os.path.dirname(abs_filename)
            try:
                object_directory = os.path.relpath(object_directory, chdir)
            except OSError:  # pragma: no cover # nosec
                pass

            out, err = gcov_cmd.run_with_args(
                [
                    abs_filename,
                    *gcov_cmd.get_default_options(),
                    "--object-directory",
                    object_directory,
                ],
                cwd=chdir,
                activate_trace_logging=not is_file_excluded(
                    "trace",
                    abs_filename,
                    options.trace_include_filter,
                    options.trace_exclude_filter,
                ),
            )

            # find the files that gcov created
            active_gcov_files, all_gcov_files = select_gcov_files_from_stdout(
                out,
                include_filter=options.gcov_include_filter,
                exclude_filter=options.gcov_exclude_filter,
                chdir=chdir,
            )
            # Remove the not used files
            remove_existing_files(list(all_gcov_files - active_gcov_files))

            ignore_source_errors = options.gcov_ignore_errors is not None and any(
                v in options.gcov_ignore_errors for v in ["all", "source_not_found"]
            )
            ignore_output_errors = options.gcov_ignore_errors is not None and any(
                v in options.gcov_ignore_errors for v in ["all", "output_error"]
            )
            if (
                # GCOV did not find source file and error shall not be ignored
                source_error_re.search(err) and not ignore_source_errors
            ):
                raise GcovMessageOnStderr(
                    "GCOV could not find source file, this can be ignored with --gcov-ignore-errors=source_not_found."
                )
            if (
                # GCOV can not write output file and error shall not be ignored
                output_error_re.search(err) and not ignore_output_errors
            ):
                raise GcovMessageOnStderr(
                    "GCOV could not write output file, this can be ignored with --gcov-ignore-errors=output_error."
                )

            if ignore_output_errors:
                active_gcov_files = set(
                    f for f in active_gcov_files if os.path.exists(f)
                )

            if options.keep_intermediate_files:
                # Keep the files with unique names
                basename = os.path.basename(abs_filename)
                renamed_active_gcov_files = set[str]()
                for gcov_filename in active_gcov_files:
                    directory, filename = os.path.split(gcov_filename)
                    new_name = os.path.join(directory, f"{basename}.{filename}")
                    renamed_active_gcov_files.add(new_name)
                    os.replace(gcov_filename, new_name)
                active_gcov_files = renamed_active_gcov_files

            # Process *.gcov files
            for gcov_filename in active_gcov_files:
                if not os.path.exists(gcov_filename):  # pragma: no cover
                    raise SanityCheckError(
                        f"Output file {gcov_filename} doesn't exist but no error from GCOV detected."
                    )
                if gcov_filename.endswith(".gcov"):
                    process_gcov_text_data(
                        gcov_filename, filename, covdata, options, chdir
                    )
                elif gcov_filename.endswith(".gcov.json.gz"):
                    process_gcov_json_data(gcov_filename, covdata, options)
                else:  # pragma: no cover
                    raise RuntimeError(f"Unknown gcov output format {gcov_filename}.")

            done = True

        except RuntimeError as exc:
            # If we got an merge assertion error we must end the processing
            done = False
            error(
                f"With working directory {chdir!r}.\n"
                f"{str(exc)}\n"
                f"STDERR >>{err}<< End of STDERR\n"
                f"STDOUT >>{out}<< End of STDOUT"
            )
        except GcovMessageOnStderr as exc:
            done = False
            error(
                f"With working directory {chdir!r}.\n"
                f"{str(exc)}\n"
                f"STDERR >>{err}<< End of STDERR"
            )
        except GcovProgram.GcovExecutionError as exc:
            done = False
            error(f"With working directory {chdir!r}.\n{str(exc)}")
        finally:
            if not (options.keep_intermediate_files and done):
                # Remove the used files
                remove_existing_files(list(active_gcov_files))

    return done


def select_gcov_files_from_stdout(
    out: str,
    include_filter: tuple[Filter, ...],
    exclude_filter: tuple[Filter, ...],
    chdir: str,
) -> tuple[set[str], set[str]]:
    """Parse the output to get the list of files to use and all files (unfiltered)."""
    active_files = set()
    all_files = set()

    for line in out.splitlines():
        found = output_re.search(line.strip())
        if found is None:
            continue

        fname = found.group(1)
        full = os.path.join(chdir, fname)
        all_files.add(full)

        if is_file_excluded("gcov file", fname, include_filter, exclude_filter):
            continue

        active_files.add(full)

    return active_files, all_files


#
#  Process Already existing gcov files
#
def process_existing_gcov_file(
    filename: str, covdata: CoverageContainer, options: Options, to_erase: set[str]
) -> None:
    """Process an existing GCOV filename."""
    if is_file_excluded(
        "gcov file", filename, options.gcov_include_filter, options.gcov_exclude_filter
    ):
        if not is_file_excluded(
            "trace",
            filename,
            options.trace_include_filter,
            options.trace_exclude_filter,
        ):
            LOGGER.trace("Excluding gcov file: %s", filename)
        return

    if filename.endswith(".gcov"):
        process_gcov_text_data(filename, None, covdata, options)
    elif filename.endswith(".gcov.json.gz"):
        process_gcov_json_data(filename, covdata, options)
    else:  # pragma: no cover
        raise RuntimeError(f"Unknown gcov output format {filename}.")

    if not options.keep_intermediate_files:
        to_erase.add(filename)



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from __future__ import annotations
from typing import TypeVar

from .merging import MergeOptions


LinecovCollectionKeyType = int
LinecovKeyType = str
BranchcovKeyType = tuple[int, int, int]
ConditioncovKeyType = tuple[int, int]
CallcovKeyType = tuple[int, int, int]
FunctioncovKeyType = str
_Key = TypeVar(
    "_Key",
    LinecovCollectionKeyType,
    LinecovKeyType,
    BranchcovKeyType,
    ConditioncovKeyType,
    CallcovKeyType,
    FunctioncovKeyType,
)
_T = TypeVar("_T")


class CoverageDict(dict[_Key, _T]):
    """Base class for a coverage dictionary."""

    def merge(
        self,
        other: CoverageDict[_Key, _T],
        options: MergeOptions,
    ) -> None:
        """Helper function to merge items in a dictionary."""

        # Merge other into self
        for key, item in other.items():
            if key in self:
                self[key].merge(item, options)
            else:
                self[key] = item

        # At this point, "self" contains all merged items.
        # The caller should not access "other" objects therefore we clear it.
        other.clear()



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from ...data_model.container import CoverageContainer
from ...formats.base import BaseHandler
from ...options import (
    GcovrConfigOption,
    GcovrDeprecatedConfigOptionAction,
    OutputOrDefault,
)


class UseLcovFormatVersion(GcovrDeprecatedConfigOptionAction):
    """Argparse action to map old option --lcov-format-v1 to new option --lcov-format-version=1.x."""

    option = "--lcov-format-version"
    config = "lcov_format_version"
    value = "1.x"


class LcovHandler(BaseHandler):
    """Class to handle LCOV format."""

    @classmethod
    def get_options(cls) -> list[GcovrConfigOption | str]:
        return [
            GcovrConfigOption(
                "lcov",
                ["--lcov"],
                group="output_options",
                metavar="OUTPUT",
                help=(
                    "Generate a LCOV info file. "
                    "OUTPUT is optional and defaults to --output."
                ),
                nargs="?",
                type=OutputOrDefault,
                default=None,
                const=OutputOrDefault(None),
            ),
            GcovrConfigOption(
                "lcov_format_version",
                ["--lcov-format-version"],
                config="lcov_format_version",
                group="output_options",
                help="The format version to write.",
                choices=("1.x", "2.0"),
                default="2.0",
            ),
            GcovrConfigOption(
                "lcov_format_version",
                ["--lcov-format-1.x"],
                group="output_options",
                help="Deprecated, please use --lcov-format-version=1.x instead.",
                nargs=0,
                action=UseLcovFormatVersion,
            ),
            GcovrConfigOption(
                "lcov_comment",
                ["--lcov-comment"],
                group="output_options",
                metavar="COMMENT",
                help="The comment used in LCOV file.",
            ),
            GcovrConfigOption(
                "lcov_test_name",
                ["--lcov-test-name"],
                group="output_options",
                metavar="NAME",
                help=(
                    "The name used for TN in LCOV file, must not contain spaces. "
                    "Default is '{default!s}'."
                ),
                default="GCOVR_report",
            ),
        ]

    def validate_options(self) -> None:
        if (
            self.options.lcov_test_name is not None
            and " " in self.options.lcov_test_name
        ):
            raise RuntimeError(
                f"The LCOV test name must not contain spaces, got {self.options.lcov_test_name!r}."
            )

    def write_report(self, covdata: CoverageContainer, output_file: str) -> None:
        from .write import write_report  # pylint: disable=import-outside-toplevel # Lazy loading is intended here

        write_report(covdata, output_file, self.options)



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

"""
Handle exclusion markers and any other source code level filtering mechanisms.

The different mechanisms are exposed as separate passes/functions
that remove unwanted aspects from the coverage data.
Alternatively, they full suite of exclusion rules can be invoked
via ``apply_all_exclusions()``, which is configured via the usual options object.
"""

from dataclasses import dataclass, field
import re

from ..data_model.coverage import FileCoverage
from ..logging import LOGGER
from ..options import Options

from .markers import ExclusionPredicate, FunctionListByLine, apply_exclusion_markers
from .noncode import remove_unreachable_branches, remove_noncode_lines
from .utils import (
    make_is_in_any_range_inclusive,
    apply_exclusion_ranges,
    function_exclude_not_supported,
    get_function_exclude_ranges,
    get_functions_by_line,
)


@dataclass
class ExclusionOptions:
    """
    Options used by exclusion processing.

    The defaults are just for testing purposes.
    Otherwise, this class acts more like an interface,
    describing some options in "gcovr.configuration".
    """

    respect_exclusion_markers: bool = True
    warn_excluded_lines_with_hits: bool = False
    exclude_function: list[re.Pattern[str]] = field(default_factory=lambda: [])
    exclude_lines_by_pattern: list[re.Pattern[str]] = field(default_factory=lambda: [])
    exclude_branches_by_pattern: list[re.Pattern[str]] = field(
        default_factory=lambda: []
    )
    exclude_pattern_prefix: str = "PREFIX"
    exclude_throw_branches: bool = False
    exclude_unreachable_branches: bool = False
    exclude_function_lines: bool = False
    exclude_internal_functions: bool = False
    exclude_noncode_lines: bool = False


def get_exclusion_options_from_options(options: Options) -> ExclusionOptions:
    """Get the exclusion options."""

    return ExclusionOptions(
        respect_exclusion_markers=options.respect_exclusion_markers,
        warn_excluded_lines_with_hits=options.warn_excluded_lines_with_hits,
        exclude_function=options.exclude_function,
        exclude_lines_by_pattern=options.exclude_lines_by_pattern,
        exclude_branches_by_pattern=options.exclude_branches_by_pattern,
        exclude_pattern_prefix=options.exclude_pattern_prefix,
        exclude_throw_branches=options.exclude_throw_branches,
        exclude_unreachable_branches=(options.exclude_unreachable_branches),
        exclude_function_lines=options.exclude_function_lines,
        exclude_internal_functions=options.exclude_internal_functions,
        exclude_noncode_lines=options.exclude_noncode_lines,
    )


def apply_all_exclusions(
    filecov: FileCoverage,
    *,
    lines: list[str],
    options: ExclusionOptions,
    activate_trace_logging: bool = False,
) -> None:
    """
    Apply all available exclusion mechanisms, if they are enabled by the options.

    Modifies the FileCoverage in place.
    """

    if options.exclude_internal_functions:
        remove_internal_functions(
            filecov, activate_trace_logging=activate_trace_logging
        )

    if options.exclude_throw_branches:
        remove_throw_branches(filecov, activate_trace_logging=activate_trace_logging)

    if options.exclude_unreachable_branches:
        remove_unreachable_branches(
            filecov, lines=lines, activate_trace_logging=activate_trace_logging
        )

    if options.exclude_noncode_lines:
        remove_noncode_lines(
            filecov, lines=lines, activate_trace_logging=activate_trace_logging
        )

    if options.respect_exclusion_markers:
        apply_exclusion_markers(
            filecov,
            lines=lines,
            exclude_lines_by_pattern=options.exclude_lines_by_pattern,
            exclude_branches_by_pattern=options.exclude_branches_by_pattern,
            exclude_pattern_prefix=options.exclude_pattern_prefix,
            warn_excluded_lines_with_hits=options.warn_excluded_lines_with_hits,
            activate_trace_logging=activate_trace_logging,
        )

    if options.exclude_function_lines:
        exclude_function_definition_lines(
            filecov, activate_trace_logging=activate_trace_logging
        )

    if options.exclude_function:
        exclude_functions(
            filecov,
            options.exclude_function,
            activate_trace_logging=activate_trace_logging,
        )


def exclude_function_definition_lines(
    filecov: FileCoverage, activate_trace_logging: bool
) -> None:
    """Remove coverage for lines that contain a function definition."""
    # iterate over a shallow copy
    known_function_lines = set(
        (lineno, functioncov.name)
        for functioncov in filecov.functioncov()
        for lineno in functioncov.linenos
    )
    for linecov in list(filecov.linecov()):
        if (linecov.lineno, linecov.function_name) in known_function_lines or (
            linecov.lineno,
            linecov.demangled_function_name,
        ) in known_function_lines:
            if activate_trace_logging:
                LOGGER.trace(
                    "%s: Removing line of function definition for function %s",
                    linecov.location,
                    linecov.function_name,
                )
            linecov.exclude()


def exclude_functions(
    filecov: FileCoverage, patterns: list[re.Pattern[str]], activate_trace_logging: bool
) -> None:
    """Remove matching functions"""
    if filecov.functioncov():
        functions_by_line: FunctionListByLine = get_functions_by_line(filecov)

        exclude_ranges = []
        for lineno, functions in functions_by_line.items():
            for functioncov in functions:
                for pattern in patterns:
                    if any(
                        pattern.fullmatch(name)
                        for name in (
                            functioncov.name,
                            functioncov.demangled_name,
                        )
                        if name is not None
                    ):
                        if (
                            functioncov.start is None
                            or functioncov.start[lineno] is None
                        ):
                            function_exclude_not_supported()
                        else:
                            exclude_ranges += get_function_exclude_ranges(
                                filecov.filename,
                                lineno,
                                functioncov.start[lineno][1]
                                + 1,  # Cheat that the comment is after the definition
                                functions_by_line=functions_by_line,
                            )
                        break
            if activate_trace_logging:
                LOGGER.trace(
                    "Exclusion range for functions from CLI in %s: %s.",
                    filecov.filename,
                    str(exclude_ranges),
                )
        exclusion_predicate: ExclusionPredicate = make_is_in_any_range_inclusive(
            exclude_ranges
        )
        apply_exclusion_ranges(
            filecov,
            line_is_excluded=exclusion_predicate,
            branch_is_excluded=exclusion_predicate,
            warn_excluded_lines_with_hits=False,
        )


def remove_internal_functions(
    filecov: FileCoverage, activate_trace_logging: bool
) -> None:
    """Remove compiler-generated functions, e.g. for static initialization."""

    # Get all the functions first because we want to remove some of them which will else result in an error.
    for functioncov in list(filecov.functioncov()):
        if _function_can_be_excluded(
            functioncov.mangled_name, functioncov.demangled_name
        ):
            if activate_trace_logging:
                LOGGER.trace(
                    "%s: Removing symbol %s detected as compiler generated functions, e.g. for static initialization",
                    functioncov.location,
                    functioncov.name,
                )
            filecov.remove_function_coverage(functioncov)


def _function_can_be_excluded(*names: str | None) -> bool:
    """Special names for construction/destruction of static objects will be ignored"""
    return any(
        name is not None
        and (name.startswith("__") or name.startswith("_GLOBAL__sub_I_"))
        for name in names
    )


def remove_throw_branches(filecov: FileCoverage, activate_trace_logging: bool) -> None:
    """Remove branches annotated as "throw"."""
    for linecov in filecov.linecov():
        # iterate over shallow copy
        branchcov_list = list(linecov.branches())
        for branchcov in branchcov_list:
            if branchcov.throw:
                linecov.remove_branch(branchcov)
        if activate_trace_logging:
            removed_items = len(branchcov_list) - len(list(linecov.branches()))
            if removed_items > 0:
                LOGGER.trace(
                    "%s: Removing %d unreachable branch(es) detected as exception-only code",
                    linecov.location,
                    removed_items,
                )



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from ...data_model.container import CoverageContainer
from ...formats.base import BaseHandler
from ...options import GcovrConfigOption, OutputOrDefault


THEMES = (
    "green",
    "blue",
)


class MarkdownHandler(BaseHandler):
    """Class to handle markdown format."""

    @classmethod
    def get_options(cls) -> list[GcovrConfigOption | str]:
        return [
            # Global options needed for report
            "medium_threshold",
            "high_threshold",
            "medium_threshold_branch",
            "high_threshold_branch",
            "medium_threshold_line",
            "high_threshold_line",
            # Local options
            GcovrConfigOption(
                "markdown",
                ["--markdown"],
                group="output_options",
                metavar="OUTPUT",
                help="Generate a markdown report. OUTPUT is optional and defaults to --output.",
                nargs="?",
                type=OutputOrDefault,
                default=None,
                const=OutputOrDefault(None),
            ),
            GcovrConfigOption(
                "markdown_summary",
                ["--markdown-summary"],
                group="output_options",
                metavar="OUTPUT",
                help=(
                    "Generate a markdown summary report. "
                    "OUTPUT is optional and defaults to --output."
                ),
                nargs="?",
                type=OutputOrDefault,
                default=None,
                const=OutputOrDefault(None),
            ),
            GcovrConfigOption(
                "markdown_theme",
                ["--markdown-theme"],
                group="output_options",
                type=str,
                choices=THEMES,
                metavar="THEME",
                help=(
                    "Override the default color theme for the Markdown report. "
                    "Default is {default!s}."
                ),
                default=THEMES[0],
            ),
            GcovrConfigOption(
                "markdown_title",
                ["--markdown-title"],
                group="output_options",
                type=str,
                metavar="TEXT",
                help=(
                    "Override the default title of the Markdown report. "
                    "Default is {default!s}."
                ),
                default="GCC Code Coverage Report",
            ),
            GcovrConfigOption(
                "markdown_heading_level",
                ["--markdown-heading-level"],
                group="output_options",
                type=int,
                metavar="INT",
                help=(
                    "Override the default heading level of the Markdown report. "
                    "This is useful if the report is embedded in another markdown file. "
                    "Default is {default!s}."
                ),
                default=1,
            ),
            GcovrConfigOption(
                "markdown_file_link",
                ["--markdown-file-link"],
                group="output_options",
                type=str,
                metavar="TEXT",
                help=(
                    "Link the files using given URL by replacing {{file}} with the current file."
                ),
                default=None,
            ),
        ]

    def validate_options(self) -> None:
        if self.options.markdown_heading_level < 1:
            raise RuntimeError("The markdown heading level must not be less than 0.")

    def write_report(self, covdata: CoverageContainer, output_file: str) -> None:
        from .write import write_report  # pylint: disable=import-outside-toplevel # Lazy loading is intended here

        write_report(covdata, output_file, self.options)

    def write_summary_report(
        self, covdata: CoverageContainer, output_file: str
    ) -> None:
        from .write import write_summary_report  # pylint: disable=import-outside-toplevel # Lazy loading is intended here

        write_summary_report(covdata, output_file, self.options)



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from typing import Any

from jinja2 import (
    Environment,
    PackageLoader,
)

from ...data_model.container import CoverageContainer
from ...data_model.stats import SummarizedStats
from ...options import Options
from ...utils import open_text_for_writing


def templates() -> Environment:
    """Get the template environment."""
    loader: PackageLoader = PackageLoader(
        "gcovr.formats.markdown",
        package_path="default",
    )

    return Environment(
        loader=loader,
        autoescape=True,
        trim_blocks=True,
        lstrip_blocks=True,
    )


def write_report(
    covdata: CoverageContainer, output_file: str, options: Options
) -> None:
    """Produce the gcovr report in markdown."""
    data = {
        "title": options.markdown_title,
        "heading_level": options.markdown_heading_level,
        "file_link": options.markdown_file_link,
        "summary": _summary_from_stats(covdata.stats, options),
    }

    sorted_keys = covdata.sort_coverage(
        sort_key=options.sort_key,
        sort_reverse=options.sort_reverse,
        by_metric="branch" if options.sort_branches else "line",
    )
    data["entries"] = list[dict[str, Any]]()
    for key in sorted_keys:
        summary = _summary_from_stats(covdata[key].stats, options)
        summary["filename"] = covdata[key].presentable_filename(options.root_filter)
        data["entries"].append(summary)

    markdown_string = templates().get_template("report_template.md.j2").render(**data)

    with open_text_for_writing(output_file, "coverage.md", encoding="utf-8") as fh:
        fh.write(markdown_string)


def write_summary_report(
    covdata: CoverageContainer, output_file: str, options: Options
) -> None:
    """Produce the gcovr summary report in markdown."""
    data = {
        "title": options.markdown_title,
        "heading_level": options.markdown_heading_level,
        "summary": _summary_from_stats(covdata.stats, options),
    }

    markdown_string = templates().get_template("report_template.md.j2").render(**data)

    with open_text_for_writing(output_file, "coverage.md", encoding="utf-8") as fh:
        fh.write(markdown_string)


def _coverage_to_badge(
    coverage: float | None,
    medium_threshold: float,
    high_threshold: float,
    theme: str,
) -> str:
    if coverage is None:
        return "⚫"
    if coverage >= high_threshold:
        return "🔵" if theme == "blue" else "🟢"
    if coverage >= medium_threshold:
        return "🟡"

    return "🔴"


def _summary_from_stats(stats: SummarizedStats, options: Options) -> dict[str, Any]:
    summary = dict[str, Any]()

    summary["line_badge"] = _coverage_to_badge(
        stats.line.percent,
        options.medium_threshold_line,
        options.high_threshold_line,
        options.markdown_theme,
    )
    summary["function_badge"] = _coverage_to_badge(
        stats.function.percent,
        options.medium_threshold,
        options.high_threshold,
        options.markdown_theme,
    )
    summary["branch_badge"] = _coverage_to_badge(
        stats.branch.percent,
        options.medium_threshold_branch,
        options.high_threshold_branch,
        options.markdown_theme,
    )
    summary["line_covered"] = stats.line.covered
    summary["line_total"] = stats.line.total
    summary["line_percent"] = stats.line.percent_or(0.0)
    summary["function_covered"] = stats.function.covered
    summary["function_total"] = stats.function.total
    summary["function_percent"] = stats.function.percent_or(0.0)
    summary["branch_covered"] = stats.branch.covered
    summary["branch_total"] = stats.branch.total
    summary["branch_percent"] = stats.branch.percent_or(0.0)

    return summary



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

import os
from typing import Any

from ...data_model import version
from ...data_model.container import CoverageContainer
from ...options import Options
from ...utils import force_unix_separator, write_json_output

SUMMARY_FORMAT_VERSION = (
    # BEGIN summary version
    "0.6"
    # END summary version
)
KEY_SUMMARY_FORMAT_VERSION = "gcovr/summary_format_version"


def write_report(
    covdata: CoverageContainer, output_file: str, options: Options
) -> None:
    """Produce an JSON report in the format partially compatible with gcov JSON output."""
    write_json_output(
        {
            "gcovr/format_version": version.FORMAT_VERSION,
            "files": covdata.serialize(options),
        },
        pretty=options.json_pretty,
        filename=output_file,
        default_filename="coverage.json",
    )


def write_summary_report(
    covdata: CoverageContainer, output_file: str, options: Options
) -> None:
    """Produce gcovr JSON summary report."""

    json_dict = dict[str, Any]()

    json_dict["root"] = force_unix_separator(
        os.path.relpath(
            options.root,
            os.getcwd() if output_file == "-" else os.path.dirname(output_file),
        )
    )
    json_dict["gcovr/summary_format_version"] = SUMMARY_FORMAT_VERSION
    files = list[dict[str, Any]]()
    json_dict["files"] = files

    # Data
    sorted_keys = covdata.sort_coverage(
        sort_key=options.sort_key,
        sort_reverse=options.sort_reverse,
        by_metric="branch" if options.sort_branches else "line",
    )

    for key in sorted_keys:
        filename = covdata[key].presentable_filename(options.root_filter)
        if options.json_base:
            filename = "/".join([options.json_base, filename])

        files.append(
            {
                "filename": filename,
                **covdata[key].stats.serialize(None, options),
            }
        )

    # Footer & summary
    json_dict.update(covdata.stats.serialize(0.0, options))

    write_json_output(
        json_dict,
        pretty=options.json_summary_pretty,
        filename=output_file,
        default_filename="summary_coverage.json",
    )
